#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import json
import os
import time
import unittest
from unittest.mock import patch, MagicMock

import evdev
from evdev._ecodes import EV_ABS
from evdev.ecodes import EV_KEY, KEY_B, KEY_A, ABS_X, BTN_A, BTN_B

from inputremapper.configs.global_config import GlobalConfig
from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.configs.mapping import Mapping
from inputremapper.configs.paths import PathUtils
from inputremapper.configs.preset import Preset
from inputremapper.configs.keyboard_layout import keyboard_layout
from inputremapper.daemon import Daemon, DAEMON
from inputremapper.groups import groups
from inputremapper.injection.global_uinputs import GlobalUInputs, UInput
from inputremapper.injection.injector import InjectorState
from inputremapper.injection.mapping_handlers.mapping_parser import MappingParser
from inputremapper.input_event import InputEvent
from tests.lib.cleanup import cleanup
from tests.lib.fixtures import Fixture
from tests.lib.fixtures import fixtures
from tests.lib.logger import logger
from tests.lib.pipes import push_events, uinput_write_history_pipe
from tests.lib.test_setup import test_setup, is_service_running
from tests.lib.tmp import tmp


@test_setup
class TestDaemon(unittest.TestCase):
    new_fixture_path = "/dev/input/event9876"

    def setUp(self):
        self.daemon = None
        self.global_config = GlobalConfig()
        self.global_uinputs = GlobalUInputs(UInput)
        PathUtils.mkdir(PathUtils.get_config_path())
        self.global_config._save_config()
        self.mapping_parser = MappingParser(self.global_uinputs)

        # the daemon should be able to create them on demand:
        self.global_uinputs.devices = {}
        self.global_uinputs.is_service = True

    def tearDown(self):
        # avoid race conditions with other tests, daemon may run processes
        if self.daemon is not None:
            self.daemon.stop_all()
            self.daemon = None

        cleanup()

    @patch.object(os, "system")
    def test_connect(self, os_system_mock: MagicMock):
        self.assertFalse(is_service_running())

        # no daemon runs, should try to run it via pkexec instead.
        # It fails due to the patch on os.system and therefore exits the process
        self.assertRaises(SystemExit, Daemon.connect)
        os_system_mock.assert_called_once()
        self.assertIsNone(Daemon.connect(False))

        # make the connect command work this time by acting like a connection is
        # available:

        set_config_dir_callcount = 0

        class FakeConnection:
            def set_config_dir(self, *_, **__):
                nonlocal set_config_dir_callcount
                set_config_dir_callcount += 1

            def Introspect(self, timeout=0):
                return ""

        with patch.object(DAEMON, "get_proxy") as get_mock:
            get_mock.return_value = FakeConnection()
            self.assertIsInstance(Daemon.connect(), FakeConnection)
            self.assertEqual(set_config_dir_callcount, 1)

            self.assertIsInstance(Daemon.connect(False), FakeConnection)
            self.assertEqual(set_config_dir_callcount, 2)

    def test_daemon(self):
        # remove the existing system mapping to force our own into it
        if os.path.exists(PathUtils.get_config_path("xmodmap.json")):
            os.remove(PathUtils.get_config_path("xmodmap.json"))

        preset_name = "foo"

        group = groups.find(name="gamepad")

        # unrelated group that shouldn't be affected at all
        group2 = groups.find(name="Bar Device")

        preset = Preset(group.get_preset_path(preset_name))
        preset.add(
            Mapping.from_combination(
                input_combination=InputCombination(
                    [InputConfig(type=EV_KEY, code=BTN_A)]
                ),
                target_uinput="keyboard",
                output_symbol="a",
            )
        )
        preset.add(
            Mapping.from_combination(
                input_combination=InputCombination(
                    [InputConfig(type=EV_ABS, code=ABS_X, analog_threshold=-1)]
                ),
                target_uinput="keyboard",
                output_symbol="b",
            )
        )
        preset.save()
        self.global_config.set_autoload_preset(group.key, preset_name)

        """Injection 1"""

        # should forward the event unchanged
        push_events(
            fixtures.gamepad,
            [InputEvent.key(BTN_B, 1, fixtures.gamepad.get_device_hash())],
        )

        self.daemon = Daemon(
            self.global_config,
            self.global_uinputs,
            self.mapping_parser,
        )

        self.assertFalse(uinput_write_history_pipe[0].poll())

        # has been cleanedUp in setUp
        self.assertNotIn("keyboard", self.global_uinputs.devices)

        logger.info(f"start injector for {group.key}")
        self.daemon.start_injecting(group.key, preset_name)

        # created on demand
        self.assertIn("keyboard", self.global_uinputs.devices)
        self.assertNotIn("gamepad", self.global_uinputs.devices)

        self.assertEqual(self.daemon.get_state(group.key), InjectorState.STARTING)
        self.assertEqual(self.daemon.get_state(group2.key), InjectorState.UNKNOWN)

        event = uinput_write_history_pipe[0].recv()
        self.assertEqual(self.daemon.get_state(group.key), InjectorState.RUNNING)
        self.assertEqual(event.type, EV_KEY)
        self.assertEqual(event.code, BTN_B)
        self.assertEqual(event.value, 1)

        logger.info(f"stopping injector for {group.key}")
        self.daemon.stop_injecting(group.key)
        time.sleep(0.2)
        self.assertEqual(self.daemon.get_state(group.key), InjectorState.STOPPED)

        try:
            self.assertFalse(uinput_write_history_pipe[0].poll())
        except AssertionError:
            print("Unexpected", uinput_write_history_pipe[0].recv())
            # possibly a duplicate write!
            raise

        """Injection 2"""
        logger.info(f"start injector for {group.key}")
        self.daemon.start_injecting(group.key, preset_name)

        time.sleep(0.1)
        # -1234 will be classified as -1 by the injector
        push_events(
            fixtures.gamepad,
            [InputEvent.abs(ABS_X, -1234, fixtures.gamepad.get_device_hash())],
        )
        time.sleep(0.1)

        self.assertTrue(uinput_write_history_pipe[0].poll())

        # the written key is a key-down event, not the original
        # event value of -1234
        event = uinput_write_history_pipe[0].recv()

        self.assertEqual(event.type, EV_KEY)
        self.assertEqual(event.code, KEY_B)
        self.assertEqual(event.value, 1)

    def test_refresh_on_start(self):
        if os.path.exists(PathUtils.get_config_path("xmodmap.json")):
            os.remove(PathUtils.get_config_path("xmodmap.json"))

        preset_name = "foo"
        key_code = 9
        group_name = "9876 name"

        # expected key of the group
        group_key = group_name

        group = groups.find(name=group_name)
        # this test only makes sense if this device is unknown yet
        self.assertIsNone(group)

        keyboard_layout.clear()
        keyboard_layout._set("a", KEY_A)

        preset = Preset(PathUtils.get_preset_path(group_name, preset_name))
        preset.add(
            Mapping.from_combination(
                InputCombination([InputConfig(type=EV_KEY, code=key_code)]),
                "keyboard",
                "a",
            )
        )

        # make the daemon load the file instead
        with open(PathUtils.get_config_path("xmodmap.json"), "w") as file:
            json.dump(keyboard_layout._mapping, file, indent=4)
        keyboard_layout.clear()

        preset.save()
        self.global_config.set_autoload_preset(group_key, preset_name)
        self.daemon = Daemon(
            self.global_config,
            self.global_uinputs,
            self.mapping_parser,
        )

        # make sure the devices are populated
        groups.refresh()

        # the daemon is supposed to find this device by calling refresh
        fixture = Fixture(
            capabilities={evdev.ecodes.EV_KEY: [key_code]},
            phys="9876 phys",
            info=evdev.device.DeviceInfo(4, 5, 6, 7),
            name=group_name,
            path=self.new_fixture_path,
        )
        fixtures[self.new_fixture_path] = fixture
        push_events(fixture, [InputEvent.key(key_code, 1, fixture.get_device_hash())])
        self.daemon.start_injecting(group_key, preset_name)

        # test if the injector called groups.refresh successfully
        group = groups.find(key=group_key)
        self.assertEqual(group.name, group_name)
        self.assertEqual(group.key, group_key)

        time.sleep(0.1)
        self.assertTrue(uinput_write_history_pipe[0].poll())

        event = uinput_write_history_pipe[0].recv()
        self.assertEqual(event, (EV_KEY, KEY_A, 1))

        self.daemon.stop_injecting(group_key)
        time.sleep(0.2)
        self.assertEqual(self.daemon.get_state(group_key), InjectorState.STOPPED)

    def test_refresh_for_unknown_key(self):
        device_9876 = "9876 name"
        # this test only makes sense if this device is unknown yet
        self.assertIsNone(groups.find(name=device_9876))

        self.daemon = Daemon(
            self.global_config,
            self.global_uinputs,
            self.mapping_parser,
        )

        # make sure the devices are populated
        groups.refresh()

        self.daemon.refresh()

        fixtures[self.new_fixture_path] = Fixture(
            capabilities={evdev.ecodes.EV_KEY: [evdev.ecodes.KEY_A]},
            phys="9876 phys",
            info=evdev.device.DeviceInfo(4, 5, 6, 7),
            name=device_9876,
            path=self.new_fixture_path,
        )

        self.daemon._autoload("25v7j9q4vtj")
        # this is unknown, so the daemon will scan the devices again

        # test if the injector called groups.refresh successfully
        self.assertIsNotNone(groups.find(name=device_9876))

    def test_xmodmap_file(self):
        """Create a custom xmodmap file, expect the daemon to read keycodes from it."""
        from_keycode = evdev.ecodes.KEY_A
        target = "keyboard"
        to_name = "q"
        to_keycode = 100

        name = "Bar Device"
        preset_name = "foo"
        group = groups.find(name=name)

        config_dir = os.path.join(tmp, "foo")

        path = os.path.join(config_dir, "presets", name, f"{preset_name}.json")

        preset = Preset(path)
        preset.add(
            Mapping.from_combination(
                InputCombination([InputConfig(type=EV_KEY, code=from_keycode)]),
                target,
                to_name,
            )
        )
        preset.save()

        keyboard_layout.clear()

        push_events(
            fixtures.bar_device,
            [
                InputEvent.key(
                    from_keycode,
                    1,
                    origin_hash=fixtures.bar_device.get_device_hash(),
                )
            ],
        )

        # an existing config file is needed otherwise set_config_dir refuses
        # to use the directory
        config_path = os.path.join(config_dir, "config.json")
        self.global_config.path = config_path
        self.global_config._save_config()

        # finally, create the xmodmap file
        xmodmap_path = os.path.join(config_dir, "xmodmap.json")
        with open(xmodmap_path, "w") as file:
            file.write(f'{{"{to_name}":{to_keycode}}}')

        # test setup complete

        self.daemon = Daemon(
            self.global_config,
            self.global_uinputs,
            self.mapping_parser,
        )
        self.daemon.set_config_dir(config_dir)

        self.daemon.start_injecting(group.key, preset_name)

        time.sleep(0.1)
        self.assertTrue(uinput_write_history_pipe[0].poll())

        event = uinput_write_history_pipe[0].recv()
        self.assertEqual(event.type, EV_KEY)
        self.assertEqual(event.code, to_keycode)
        self.assertEqual(event.value, 1)

    def test_start_stop(self):
        group_key = "Qux/[Device]?"
        group = groups.find(key=group_key)
        preset_name = "preset8"

        daemon = Daemon(
            self.global_config,
            self.global_uinputs,
            self.mapping_parser,
        )
        self.daemon = daemon

        pereset = Preset(group.get_preset_path(preset_name))
        pereset.add(
            Mapping.from_combination(
                InputCombination([InputConfig(type=EV_KEY, code=KEY_A)]),
                "keyboard",
                "a",
            )
        )
        pereset.save()

        # start
        daemon.start_injecting(group_key, preset_name)
        # explicit start, not autoload, so the history stays empty
        self.assertNotIn(group_key, daemon.autoload_history._autoload_history)
        self.assertTrue(daemon.autoload_history.may_autoload(group_key, preset_name))
        # path got translated to the device name
        self.assertIn(group_key, daemon.injectors)

        # start again
        previous_injector = daemon.injectors[group_key]
        self.assertNotEqual(previous_injector.get_state(), InjectorState.STOPPED)
        daemon.start_injecting(group_key, preset_name)
        self.assertNotIn(group_key, daemon.autoload_history._autoload_history)
        self.assertTrue(daemon.autoload_history.may_autoload(group_key, preset_name))
        self.assertIn(group_key, daemon.injectors)
        time.sleep(0.2)
        self.assertEqual(previous_injector.get_state(), InjectorState.STOPPED)
        # a different injetor is now running
        self.assertNotEqual(previous_injector, daemon.injectors[group_key])
        self.assertNotEqual(
            daemon.injectors[group_key].get_state(), InjectorState.STOPPED
        )

        # trying to inject a non existing preset keeps the previous inejction
        # alive
        injector = daemon.injectors[group_key]
        daemon.start_injecting(group_key, "qux")
        self.assertEqual(injector, daemon.injectors[group_key])
        self.assertNotEqual(
            daemon.injectors[group_key].get_state(), InjectorState.STOPPED
        )

        # trying to start injecting for an unknown device also just does
        # nothing
        daemon.start_injecting("quux", "qux")
        self.assertNotEqual(
            daemon.injectors[group_key].get_state(), InjectorState.STOPPED
        )

        # after all that stuff autoload_history is still unharmed
        self.assertNotIn(group_key, daemon.autoload_history._autoload_history)
        self.assertTrue(daemon.autoload_history.may_autoload(group_key, preset_name))

        # stop
        daemon.stop_injecting(group_key)
        time.sleep(0.2)
        self.assertNotIn(group_key, daemon.autoload_history._autoload_history)
        self.assertEqual(daemon.injectors[group_key].get_state(), InjectorState.STOPPED)
        self.assertTrue(daemon.autoload_history.may_autoload(group_key, preset_name))

    def test_autoload(self):
        preset_name = "preset7"
        group_key = "Qux/[Device]?"
        group = groups.find(key=group_key)

        daemon = Daemon(self.global_config, self.global_uinputs, self.mapping_parser)
        self.daemon = daemon

        preset = Preset(group.get_preset_path(preset_name))
        preset.add(
            Mapping.from_combination(
                InputCombination([InputConfig(type=EV_KEY, code=KEY_A)]),
                "keyboard",
                "a",
            )
        )
        preset.save()

        # no autoloading is configured yet
        self.daemon._autoload(group_key)
        self.assertNotIn(group_key, daemon.autoload_history._autoload_history)
        self.assertTrue(daemon.autoload_history.may_autoload(group_key, preset_name))

        self.global_config.set_autoload_preset(group_key, preset_name)
        len_before = len(self.daemon.autoload_history._autoload_history)
        # now autoloading is configured, so it will autoload
        self.daemon._autoload(group_key)
        len_after = len(self.daemon.autoload_history._autoload_history)
        self.assertEqual(
            daemon.autoload_history._autoload_history[group_key][1],
            preset_name,
        )
        self.assertFalse(daemon.autoload_history.may_autoload(group_key, preset_name))
        injector = daemon.injectors[group_key]
        self.assertEqual(len_before + 1, len_after)

        # calling duplicate get_autoload does nothing
        self.daemon._autoload(group_key)
        self.assertEqual(
            daemon.autoload_history._autoload_history[group_key][1],
            preset_name,
        )
        self.assertEqual(injector, daemon.injectors[group_key])
        self.assertFalse(daemon.autoload_history.may_autoload(group_key, preset_name))

        # explicit start_injecting clears the autoload history
        self.daemon.start_injecting(group_key, preset_name)
        self.assertTrue(daemon.autoload_history.may_autoload(group_key, preset_name))

        # calling autoload for (yet) unknown devices does nothing
        len_before = len(self.daemon.autoload_history._autoload_history)
        self.daemon._autoload("unknown-key-1234")
        len_after = len(self.daemon.autoload_history._autoload_history)
        self.assertEqual(len_before, len_after)

        # autoloading input-remapper devices does nothing
        len_before = len(self.daemon.autoload_history._autoload_history)
        self.daemon.autoload_single("Bar Device")
        len_after = len(self.daemon.autoload_history._autoload_history)
        self.assertEqual(len_before, len_after)

    def test_autoload_2(self):
        self.daemon = Daemon(
            self.global_config,
            self.global_uinputs,
            self.mapping_parser,
        )
        history = self.daemon.autoload_history._autoload_history

        # existing device
        preset_name = "preset7"
        group = groups.find(key="Foo Device 2")
        preset = Preset(group.get_preset_path(preset_name))
        preset.add(
            Mapping.from_combination(
                InputCombination([InputConfig(type=3, code=2, analog_threshold=1)]),
                "keyboard",
                "a",
            )
        )
        preset.save()
        self.global_config.set_autoload_preset(group.key, preset_name)

        # ignored, won't cause problems:
        self.global_config.set_autoload_preset("non-existant-key", "foo")

        self.daemon.autoload()
        self.assertEqual(len(history), 1)
        self.assertEqual(history[group.key][1], preset_name)

    def test_autoload_3(self):
        # based on a bug
        preset_name = "preset7"
        group = groups.find(key="Foo Device 2")

        preset = Preset(group.get_preset_path(preset_name))
        preset.add(
            Mapping.from_combination(
                InputCombination([InputConfig(type=3, code=2, analog_threshold=1)]),
                "keyboard",
                "a",
            )
        )
        preset.save()

        self.global_config.set_autoload_preset(group.key, preset_name)

        self.daemon = Daemon(
            self.global_config,
            self.global_uinputs,
            self.mapping_parser,
        )
        groups.set_groups([])  # caused the bug
        self.assertIsNone(groups.find(key="Foo Device 2"))
        self.daemon.autoload()

        # it should try to refresh the groups because all the
        # group_keys are unknown at the moment
        history = self.daemon.autoload_history._autoload_history
        self.assertEqual(history[group.key][1], preset_name)
        self.assertEqual(self.daemon.get_state(group.key), InjectorState.STARTING)
        self.assertIsNotNone(groups.find(key="Foo Device 2"))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import itertools
from typing import Tuple, Iterable, Union, List, Dict, Optional, Hashable

from evdev import ecodes
from inputremapper.configs.paths import PathUtils
from inputremapper.input_event import InputEvent

try:
    from pydantic.v1 import BaseModel, root_validator, validator
except ImportError:
    from pydantic import BaseModel, root_validator, validator

from inputremapper.configs.keyboard_layout import keyboard_layout
from inputremapper.gui.messages.message_types import MessageType
from inputremapper.logging.logger import logger
from inputremapper.utils import get_evdev_constant_name, DeviceHash

# having shift in combinations modifies the configured output,
# ctrl might not work at all
DIFFICULT_COMBINATIONS = [
    ecodes.KEY_LEFTSHIFT,
    ecodes.KEY_RIGHTSHIFT,
    ecodes.KEY_LEFTCTRL,
    ecodes.KEY_RIGHTCTRL,
    ecodes.KEY_LEFTALT,
    ecodes.KEY_RIGHTALT,
]

EMPTY_TYPE = 99

# "Button 11" (10 + 0x110 "BTN_LEFT") is the highest mouse button simulatable
# at time of writing using Piper 0.8 and a Logi G502X.
# Please update if a way to simulate higher button-presses is found.
MAX_BTN_MOUSE_ECODE = 0x11A


class InputConfig(BaseModel):
    """Describes a single input within a combination, to configure mappings."""

    message_type = MessageType.selected_event

    type: int
    code: int

    # origin_hash is a hash to identify a specific /dev/input/eventXX device.
    # This solves a number of bugs when multiple devices have overlapping capabilities.
    # see utils.get_device_hash for the exact hashing function
    origin_hash: Optional[DeviceHash] = None

    # At which point is an analog input treated as "pressed"
    analog_threshold: Optional[int] = None

    def __str__(self):
        return f"InputConfig {get_evdev_constant_name(self.type, self.code)}"

    def __repr__(self):
        return (
            f"<InputConfig {self.type_and_code} "
            f"{get_evdev_constant_name(*self.type_and_code)}, "
            f"{self.analog_threshold}, "
            f"{self.origin_hash}, "
            f"at {hex(id(self))}>"
        )

    @property
    def input_match_hash(self) -> Hashable:
        """a Hashable object which is intended to match the InputConfig with a
        InputEvent.

        InputConfig itself is hashable, but can not be used to match InputEvent's
        because its hash includes the analog_threshold
        """
        return self.type, self.code, self.origin_hash

    @property
    def is_empty(self) -> bool:
        return self.type == EMPTY_TYPE

    @property
    def defines_analog_input(self) -> bool:
        """Whether this defines an analog input."""
        return not self.analog_threshold and self.type != ecodes.EV_KEY

    @property
    def type_and_code(self) -> Tuple[int, int]:
        """Event type, code."""
        return self.type, self.code

    @classmethod
    def btn_left(cls):
        return cls(type=ecodes.EV_KEY, code=ecodes.BTN_LEFT)

    @classmethod
    def from_input_event(cls, event: InputEvent) -> InputConfig:
        """create an input confing from the given InputEvent, uses the value as
        analog threshold"""
        return cls(
            type=event.type,
            code=event.code,
            origin_hash=event.origin_hash,
            analog_threshold=event.value,
        )

    def description(self, exclude_threshold=False, exclude_direction=False) -> str:
        """Get a human-readable description of the event."""
        return (
            f"{self._get_name()} "
            f"{self._get_direction() if not exclude_direction else ''} "
            f"{self._get_threshold_value() if not exclude_threshold else ''}".strip()
        )

    def _get_mouse_button_name(self) -> Optional[str]:
        """Get a human-readable description of a mouse-button. Only the first 7
        mouse buttons are in evdev and they often have misleading names there
        (eg it calls buttons 6 & 7 forward/back but usually that's buttons 5 & 4).
        Returns None if not a mouse button."""

        if self.type == ecodes.EV_KEY:
            if ecodes.BTN_MOUSE <= self.code <= ecodes.BTN_MIDDLE:
                # button is left/right/middle button
                key_name: str = get_evdev_constant_name(self.type, self.code)
                return key_name.replace(
                    "BTN_", "Mouse Button "
                )  # eg "Mouse Button LEFT"
            elif ecodes.BTN_MIDDLE < self.code <= MAX_BTN_MOUSE_ECODE:
                # button is a higher-number mouse button like side-buttons.
                # This calculation assumes left mouse button is button 1, so side buttons start at 4.
                button_number: int = self.code - ecodes.BTN_MOUSE + 1
                return f"Mouse Button {button_number}"  # eg "Mouse Button 7"

        return None

    def _get_name(self) -> Optional[str]:
        """Human-readable name (e.g. KEY_A) of the specified input event."""

        # prevent logging warnings for new/empty configs
        if self.is_empty:
            return None

        # must check if it's a mouse button *before* ecodes
        # because not all mouse buttons are in ecodes.
        mouse_button_name: Optional[str] = self._get_mouse_button_name()
        if mouse_button_name is not None:
            return mouse_button_name

        if self.type not in ecodes.bytype:
            logger.warning("Unknown type for %s", self)
            return f"unknown {self.type, self.code}"

        if self.code not in ecodes.bytype[self.type]:
            logger.warning("Unknown code for %s", self)
            return f"unknown {self.type, self.code}"

        key_name = None

        # first try to find the name in xmodmap to not display wrong
        # names due to the keyboard layout
        if self.type == ecodes.EV_KEY:
            key_name = keyboard_layout.get_name(self.code)

        if key_name is None:
            # if no result, look in the linux combination constants. On a german
            # keyboard for example z and y are switched, which will therefore
            # cause the wrong letter to be displayed.
            key_name = get_evdev_constant_name(self.type, self.code)

        key_name = key_name.replace("ABS_Z", "Trigger Left")
        key_name = key_name.replace("ABS_RZ", "Trigger Right")

        key_name = key_name.replace("ABS_HAT0X", "DPad-X")
        key_name = key_name.replace("ABS_HAT0Y", "DPad-Y")
        key_name = key_name.replace("ABS_HAT1X", "DPad-2-X")
        key_name = key_name.replace("ABS_HAT1Y", "DPad-2-Y")
        key_name = key_name.replace("ABS_HAT2X", "DPad-3-X")
        key_name = key_name.replace("ABS_HAT2Y", "DPad-3-Y")

        key_name = key_name.replace("ABS_X", "Joystick-X")
        key_name = key_name.replace("ABS_Y", "Joystick-Y")
        key_name = key_name.replace("ABS_RX", "Joystick-RX")
        key_name = key_name.replace("ABS_RY", "Joystick-RY")

        key_name = key_name.replace("BTN_", "Button ")
        key_name = key_name.replace("KEY_", "")

        key_name = key_name.replace("REL_", "")
        key_name = key_name.replace("HWHEEL", "Wheel")
        key_name = key_name.replace("WHEEL", "Wheel")

        key_name = key_name.replace("_", " ")
        key_name = key_name.replace("  ", " ")
        return key_name

    def _get_direction(self) -> str:
        """human-readable direction description for the analog_threshold"""
        if self.type == ecodes.EV_KEY or self.defines_analog_input:
            return ""

        assert self.analog_threshold
        threshold_direction = self.analog_threshold // abs(self.analog_threshold)
        return {
            # D-Pad
            (ecodes.ABS_HAT0X, -1): "Left",
            (ecodes.ABS_HAT0X, 1): "Right",
            (ecodes.ABS_HAT0Y, -1): "Up",
            (ecodes.ABS_HAT0Y, 1): "Down",
            (ecodes.ABS_HAT1X, -1): "Left",
            (ecodes.ABS_HAT1X, 1): "Right",
            (ecodes.ABS_HAT1Y, -1): "Up",
            (ecodes.ABS_HAT1Y, 1): "Down",
            (ecodes.ABS_HAT2X, -1): "Left",
            (ecodes.ABS_HAT2X, 1): "Right",
            (ecodes.ABS_HAT2Y, -1): "Up",
            (ecodes.ABS_HAT2Y, 1): "Down",
            # joystick
            (ecodes.ABS_X, 1): "Right",
            (ecodes.ABS_X, -1): "Left",
            (ecodes.ABS_Y, 1): "Down",
            (ecodes.ABS_Y, -1): "Up",
            (ecodes.ABS_RX, 1): "Right",
            (ecodes.ABS_RX, -1): "Left",
            (ecodes.ABS_RY, 1): "Down",
            (ecodes.ABS_RY, -1): "Up",
            # wheel
            (ecodes.REL_WHEEL, -1): "Down",
            (ecodes.REL_WHEEL, 1): "Up",
            (ecodes.REL_HWHEEL, -1): "Left",
            (ecodes.REL_HWHEEL, 1): "Right",
        }.get((self.code, threshold_direction)) or (
            "+" if threshold_direction > 0 else "-"
        )

    def _get_threshold_value(self) -> str:
        """human-readable value of the analog_threshold e.g. '20%'"""
        if self.analog_threshold is None:
            return ""
        return {
            ecodes.EV_REL: f"{abs(self.analog_threshold)}",
            ecodes.EV_ABS: f"{abs(self.analog_threshold)}%",
        }.get(self.type) or ""

    def modify(
        self,
        type_: Optional[int] = None,
        code: Optional[int] = None,
        origin_hash: Optional[str] = None,
        analog_threshold: Optional[int] = None,
    ) -> InputConfig:
        """Return a new modified event."""
        return InputConfig(
            type=type_ if type_ is not None else self.type,
            code=code if code is not None else self.code,
            origin_hash=origin_hash if origin_hash is not None else self.origin_hash,
            analog_threshold=(
                analog_threshold
                if analog_threshold is not None
                else self.analog_threshold
            ),
        )

    def __hash__(self):
        return hash((self.type, self.code, self.origin_hash, self.analog_threshold))

    @validator("analog_threshold")
    def _ensure_analog_threshold_is_none(cls, analog_threshold):
        """ensure the analog threshold is none, not zero."""
        if analog_threshold == 0 or analog_threshold is None:
            return None

        return analog_threshold

    @root_validator
    def _remove_analog_threshold_for_key_input(cls, values):
        """remove the analog threshold if the type is a EV_KEY"""
        type_ = values.get("type")
        if type_ == ecodes.EV_KEY:
            values["analog_threshold"] = None
        return values

    @root_validator(pre=True)
    def validate_origin_hash(cls, values):
        origin_hash = values.get("origin_hash")
        if origin_hash is None:
            # For new presets, origin_hash should be set. For old ones, it can
            # be still missing. A lot of tests didn't set an origin_hash.
            if values.get("type") != EMPTY_TYPE:
                logger.warning("No origin_hash set for %s", values)

            return values

        values["origin_hash"] = origin_hash.lower()
        return values

    class Config:
        allow_mutation = False
        underscore_attrs_are_private = True


InputCombinationInit = Union[
    Iterable[Dict[str, Union[str, int]]],
    Iterable[InputConfig],
]


class InputCombination(Tuple[InputConfig, ...]):
    """One or more InputConfigs used to trigger a mapping."""

    # tuple is immutable, therefore we need to override __new__()
    # https://jfine-python-classes.readthedocs.io/en/latest/subclass-tuple.html
    def __new__(cls, configs: InputCombinationInit) -> InputCombination:
        """Create a new InputCombination.

        Examples
        --------
            InputCombination([InputConfig, ...])
            InputCombination([{type: ..., code: ..., value: ...}, ...])
        """
        if not isinstance(configs, Iterable):
            raise TypeError("InputCombination requires a list of InputConfigs.")

        if isinstance(configs, InputConfig):
            # wrap the argument in square brackets
            raise TypeError("InputCombination requires a list of InputConfigs.")

        validated_configs = []
        for config in configs:
            if isinstance(configs, InputEvent):
                raise TypeError("InputCombinations require InputConfigs, not Events.")

            if isinstance(config, InputConfig):
                validated_configs.append(config)
            elif isinstance(config, dict):
                validated_configs.append(InputConfig(**config))
            else:
                raise TypeError(f'Can\'t handle "{config}"')

        if len(validated_configs) == 0:
            raise ValueError(f"failed to create InputCombination with {configs = }")

        # mypy bug: https://github.com/python/mypy/issues/8957
        # https://github.com/python/mypy/issues/8541
        return super().__new__(cls, validated_configs)  # type: ignore

    def __str__(self):
        return f'Combination ({" + ".join(str(event) for event in self)})'

    def __repr__(self):
        combination = ", ".join(repr(event) for event in self)
        return f"<InputCombination ({combination}) at {hex(id(self))}>"

    @classmethod
    def __get_validators__(cls):
        """Used by pydantic to create InputCombination objects."""
        yield cls.validate

    @classmethod
    def validate(cls, init_arg) -> InputCombination:
        """The only valid option is from_config"""
        if isinstance(init_arg, InputCombination):
            return init_arg
        return cls(init_arg)

    def to_config(self) -> Tuple[Dict[str, int], ...]:
        """Turn the object into a tuple of dicts."""
        return tuple(input_config.dict(exclude_defaults=True) for input_config in self)

    @classmethod
    def empty_combination(cls) -> InputCombination:
        """A combination that has default invalid (to evdev) values.

        Useful for the UI to indicate that this combination is not set
        """
        return cls([{"type": EMPTY_TYPE, "code": 99, "analog_threshold": 99}])

    @classmethod
    def from_tuples(cls, *tuples):
        """Construct an InputCombination from (type, code, analog_threshold) tuples."""
        dicts = []
        for tuple_ in tuples:
            if len(tuple_) == 3:
                dicts.append(
                    {
                        "type": tuple_[0],
                        "code": tuple_[1],
                        "analog_threshold": tuple_[2],
                    }
                )
            elif len(tuple_) == 2:
                dicts.append(
                    {
                        "type": tuple_[0],
                        "code": tuple_[1],
                    }
                )
            else:
                raise TypeError

        return cls(dicts)

    def is_problematic(self) -> bool:
        """Is this combination going to work properly on all systems?"""
        if len(self) <= 1:
            return False

        for input_config in self:
            if input_config.type != ecodes.EV_KEY:
                continue

            if input_config.code in DIFFICULT_COMBINATIONS:
                return True

        return False

    @property
    def defines_analog_input(self) -> bool:
        """Check if there is any analog input in self."""
        return True in tuple(i.defines_analog_input for i in self)

    def find_analog_input_config(
        self, type_: Optional[int] = None
    ) -> Optional[InputConfig]:
        """Return the first event that defines an analog input."""
        for input_config in self:
            if input_config.defines_analog_input and (
                type_ is None or input_config.type == type_
            ):
                return input_config
        return None

    def get_permutations(self) -> List[InputCombination]:
        """Get a list of EventCombinations representing all possible permutations.

        combining a + b + c should have the same result as b + a + c.
        Only the last combination remains the same in the returned result.
        """
        if len(self) <= 2:
            return [self]

        if len(self) > 6:
            logger.warning(
                "Your input combination has a length of %d. Long combinations might "
                'freeze the process. Edit the configuration files in "%s" to fix it.',
                len(self),
                PathUtils.get_config_path(),
            )

        permutations = []
        for permutation in itertools.permutations(self[:-1]):
            permutations.append(InputCombination((*permutation, self[-1])))

        return permutations

    def beautify(self) -> str:
        """Get a human-readable string representation."""
        if self == InputCombination.empty_combination():
            return "empty_combination"
        return " + ".join(event.description(exclude_threshold=True) for event in self)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import unittest
from functools import partial

from evdev.ecodes import (
    EV_REL,
    REL_X,
    EV_KEY,
    REL_Y,
    REL_WHEEL,
    REL_WHEEL_HI_RES,
    KEY_1,
    KEY_ESC,
)

try:
    from pydantic.v1 import ValidationError
except ImportError:
    from pydantic import ValidationError

from inputremapper.configs.mapping import Mapping, UIMapping, MappingType
from inputremapper.configs.keyboard_layout import keyboard_layout, DISABLE_NAME
from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.gui.messages.message_broker import MessageType
from tests.lib.test_setup import test_setup


@test_setup
class TestMapping(unittest.IsolatedAsyncioTestCase):
    def test_init(self):
        """Test init and that defaults are set."""
        cfg = {
            "input_combination": [{"type": 1, "code": 2}],
            "target_uinput": "keyboard",
            "output_symbol": "a",
        }
        m = Mapping(**cfg)
        self.assertEqual(
            m.input_combination, InputCombination([InputConfig(type=1, code=2)])
        )
        self.assertEqual(m.target_uinput, "keyboard")
        self.assertEqual(m.output_symbol, "a")

        self.assertIsNone(m.output_code)
        self.assertIsNone(m.output_type)

        self.assertEqual(m.macro_key_sleep_ms, 0)
        self.assertEqual(m.deadzone, 0.1)
        self.assertEqual(m.gain, 1)
        self.assertEqual(m.expo, 0)
        self.assertEqual(m.rel_rate, 60)
        self.assertEqual(m.rel_to_abs_input_cutoff, 2)
        self.assertEqual(m.release_timeout, 0.05)

    def test_is_wheel_output(self):
        mapping = Mapping(
            input_combination=InputCombination([InputConfig(type=EV_REL, code=REL_X)]),
            target_uinput="keyboard",
            output_type=EV_REL,
            output_code=REL_Y,
        )
        self.assertFalse(mapping.is_wheel_output())
        self.assertFalse(mapping.is_high_res_wheel_output())

        mapping = Mapping(
            input_combination=InputCombination([InputConfig(type=EV_REL, code=REL_X)]),
            target_uinput="keyboard",
            output_type=EV_REL,
            output_code=REL_WHEEL,
        )
        self.assertTrue(mapping.is_wheel_output())
        self.assertFalse(mapping.is_high_res_wheel_output())

        mapping = Mapping(
            input_combination=InputCombination([InputConfig(type=EV_REL, code=REL_X)]),
            target_uinput="keyboard",
            output_type=EV_REL,
            output_code=REL_WHEEL_HI_RES,
        )
        self.assertFalse(mapping.is_wheel_output())
        self.assertTrue(mapping.is_high_res_wheel_output())

    def test_get_output_type_code(self):
        cfg = {
            "input_combination": [{"type": 1, "code": 2}],
            "target_uinput": "keyboard",
            "output_symbol": "a",
        }
        m = Mapping(**cfg)
        a = keyboard_layout.get("a")
        self.assertEqual(m.get_output_type_code(), (EV_KEY, a))

        m.output_symbol = "key(a)"
        self.assertIsNone(m.get_output_type_code())

        cfg = {
            "input_combination": [{"type": 1, "code": 2}, {"type": 3, "code": 1}],
            "target_uinput": "keyboard",
            "output_type": 2,
            "output_code": 3,
        }
        m = Mapping(**cfg)
        self.assertEqual(m.get_output_type_code(), (2, 3))

    def test_strips_output_symbol(self):
        cfg = {
            "input_combination": [{"type": 1, "code": 2}],
            "target_uinput": "keyboard",
            "output_symbol": "\t a \n",
        }
        m = Mapping(**cfg)
        a = keyboard_layout.get("a")
        self.assertEqual(m.get_output_type_code(), (EV_KEY, a))

    def test_combination_changed_callback(self):
        cfg = {
            "input_combination": [{"type": 1, "code": 1}],
            "target_uinput": "keyboard",
            "output_symbol": "a",
        }
        m = Mapping(**cfg)
        arguments = []

        def callback(*args):
            arguments.append(tuple(args))

        m.set_combination_changed_callback(callback)
        m.input_combination = [{"type": 1, "code": 2}]
        m.input_combination = [{"type": 1, "code": 3}]

        # make sure a copy works as expected and keeps the callback
        m2 = m.copy()
        m2.input_combination = [{"type": 1, "code": 4}]
        m2.remove_combination_changed_callback()
        m.remove_combination_changed_callback()
        m.input_combination = [{"type": 1, "code": 5}]
        m2.input_combination = [{"type": 1, "code": 6}]
        self.assertEqual(
            arguments,
            [
                (
                    InputCombination([{"type": 1, "code": 2}]),
                    InputCombination([{"type": 1, "code": 1}]),
                ),
                (
                    InputCombination([{"type": 1, "code": 3}]),
                    InputCombination([{"type": 1, "code": 2}]),
                ),
                (
                    InputCombination([{"type": 1, "code": 4}]),
                    InputCombination([{"type": 1, "code": 3}]),
                ),
            ],
        )
        m.remove_combination_changed_callback()

    def test_init_fails(self):
        """Test that the init fails with invalid data."""
        test = partial(self.assertRaises, ValidationError, Mapping)
        cfg = {
            "input_combination": [{"type": 1, "code": 2}],
            "target_uinput": "keyboard",
            "output_symbol": "a",
        }
        Mapping(**cfg)

        # missing output symbol
        del cfg["output_symbol"]
        test(**cfg)
        cfg["output_code"] = KEY_ESC
        test(**cfg)
        cfg["output_type"] = EV_KEY
        Mapping(**cfg)

        # matching type, code and symbol. This cannot be done via the ui, and requires
        # manual editing of the preset file.
        a = keyboard_layout.get("a")
        cfg["output_code"] = a
        cfg["output_symbol"] = "a"
        cfg["output_type"] = EV_KEY
        Mapping(**cfg)

        # macro
        cfg["output_symbol"] = "key(a)"
        test(**cfg)
        cfg["output_symbol"] = "a"
        Mapping(**cfg)

        # mismatching type, code and symbol
        cfg["output_symbol"] = "b"
        test(**cfg)
        del cfg["output_type"]
        del cfg["output_code"]
        Mapping(**cfg)  # no error

        # empty symbol string without type and code
        cfg["output_symbol"] = ""
        test(**cfg)
        cfg["output_symbol"] = "a"

        # missing target
        del cfg["target_uinput"]
        test(**cfg)

        # unknown target
        cfg["target_uinput"] = "foo"
        test(**cfg)
        cfg["target_uinput"] = "keyboard"
        Mapping(**cfg)

        # missing input_combination
        del cfg["input_combination"]
        test(**cfg)
        cfg["input_combination"] = [{"type": 1, "code": 2}]
        Mapping(**cfg)

        # no macro and not a known symbol
        cfg["output_symbol"] = "qux"
        test(**cfg)
        cfg["output_symbol"] = "key(a)"
        Mapping(**cfg)

        # invalid macro
        cfg["output_symbol"] = "key('a')"
        test(**cfg)
        cfg["output_symbol"] = "a"
        Mapping(**cfg)

        # map axis but no output type and code given
        cfg["input_combination"] = [{"type": 3, "code": 0}]
        test(**cfg)
        # output symbol=disable is allowed
        cfg["output_symbol"] = DISABLE_NAME
        Mapping(**cfg)
        del cfg["output_symbol"]
        cfg["output_code"] = 1
        cfg["output_type"] = 3
        Mapping(**cfg)

        # empty symbol string is allowed when type and code is set
        cfg["output_symbol"] = ""
        Mapping(**cfg)
        del cfg["output_symbol"]

        # multiple axis as axis in event combination
        cfg["input_combination"] = [{"type": 3, "code": 0}, {"type": 3, "code": 1}]
        test(**cfg)
        cfg["input_combination"] = [{"type": 3, "code": 0}]
        Mapping(**cfg)

        del cfg["output_type"]
        del cfg["output_code"]
        cfg["input_combination"] = [{"type": 1, "code": 2}]
        cfg["output_symbol"] = "a"
        Mapping(**cfg)

        # map EV_ABS as key with trigger point out of range
        cfg["input_combination"] = [{"type": 3, "code": 0, "analog_threshold": 100}]
        test(**cfg)
        cfg["input_combination"] = [{"type": 3, "code": 0, "analog_threshold": 99}]
        Mapping(**cfg)
        cfg["input_combination"] = [{"type": 3, "code": 0, "analog_threshold": -100}]
        test(**cfg)
        cfg["input_combination"] = [{"type": 3, "code": 0, "analog_threshold": -99}]
        Mapping(**cfg)

        cfg["input_combination"] = [{"type": 1, "code": 2}]
        Mapping(**cfg)

        # deadzone out of range
        test(**cfg, deadzone=1.01)
        test(**cfg, deadzone=-0.01)
        Mapping(**cfg, deadzone=1)
        Mapping(**cfg, deadzone=0)

        # expo out of range
        test(**cfg, expo=1.01)
        test(**cfg, expo=-1.01)
        Mapping(**cfg, expo=1)
        Mapping(**cfg, expo=-1)

        # negative rate
        test(**cfg, rel_rate=-1)
        test(**cfg, rel_rate=0)

        Mapping(**cfg, rel_rate=1)
        Mapping(**cfg, rel_rate=200)

        # negative rel_to_abs_input_cutoff
        test(**cfg, rel_to_abs_input_cutoff=-1)
        test(**cfg, rel_to_abs_input_cutoff=0)
        Mapping(**cfg, rel_to_abs_input_cutoff=1)
        Mapping(**cfg, rel_to_abs_input_cutoff=3)

        # negative release timeout
        test(**cfg, release_timeout=-0.1)
        test(**cfg, release_timeout=0)
        Mapping(**cfg, release_timeout=0.05)
        Mapping(**cfg, release_timeout=0.3)

        # analog output but no analog input
        cfg = {
            "input_combination": [{"type": 3, "code": 1, "analog_threshold": -1}],
            "target_uinput": "gamepad",
            "output_type": 3,
            "output_code": 1,
        }
        test(**cfg)
        cfg["input_combination"] = [{"type": 2, "code": 1, "analog_threshold": -1}]
        test(**cfg)
        cfg["output_type"] = 2
        test(**cfg)
        cfg["input_combination"] = [{"type": 3, "code": 1, "analog_threshold": -1}]
        test(**cfg)

    def test_automatically_detects_mapping_type(self):
        cfg = {
            "input_combination": [{"type": 1, "code": 2}],
            "target_uinput": "keyboard",
            "output_symbol": "a",
        }
        self.assertEqual(Mapping(**cfg).mapping_type, MappingType.KEY_MACRO.value)

        cfg = {
            "input_combination": [{"type": 1, "code": 2}],
            "target_uinput": "keyboard",
            "output_type": EV_KEY,
            "output_code": KEY_ESC,
        }
        self.assertEqual(Mapping(**cfg).mapping_type, MappingType.KEY_MACRO.value)

        cfg = {
            "input_combination": [{"type": EV_REL, "code": REL_X}],
            "target_uinput": "keyboard",
            "output_type": EV_REL,
            "output_code": REL_X,
        }
        self.assertEqual(Mapping(**cfg).mapping_type, MappingType.ANALOG.value)

    def test_revalidate_at_assignment(self):
        cfg = {
            "input_combination": [{"type": 1, "code": 1}],
            "target_uinput": "keyboard",
            "output_symbol": "a",
        }
        m = Mapping(**cfg)
        test = partial(self.assertRaises, ValidationError, m.__setattr__)

        # invalid input event
        test("input_combination", "1,2,3,4")

        # unknown target
        test("target_uinput", "foo")

        # invalid macro
        test("output_symbol", "key()")

        # we could do a lot more tests here but since pydantic uses the same validation
        # code as for the initialization we only need to make sure that the
        # assignment validation is active

    def test_set_invalid_combination_with_callback(self):
        cfg = {
            "input_combination": [{"type": 1, "code": 1}],
            "target_uinput": "keyboard",
            "output_symbol": "a",
        }
        m = Mapping(**cfg)
        m.set_combination_changed_callback(lambda *args: None)
        self.assertRaises(ValidationError, m.__setattr__, "input_combination", "1,2")
        m.input_combination = [{"type": 1, "code": 2}]
        m.input_combination = [{"type": 1, "code": 2}]

    def test_is_valid(self):
        cfg = {
            "input_combination": [{"type": 1, "code": 1}],
            "target_uinput": "keyboard",
            "output_symbol": "a",
        }
        m = Mapping(**cfg)
        self.assertTrue(m.is_valid())

    def test_wrong_target(self):
        mapping = Mapping(
            input_combination=[{"type": EV_KEY, "code": KEY_1}],
            target_uinput="keyboard",
            output_symbol="a",
        )
        mapping.set_combination_changed_callback(lambda *args: None)
        self.assertRaisesRegex(
            ValidationError,
            # the error should mention
            # - the symbol
            # - the current incorrect target
            # - the target that works for this symbol
            ".*BTN_A.*keyboard.*gamepad",
            mapping.__setattr__,
            "output_symbol",
            "BTN_A",
        )

    def test_wrong_target_for_macro(self):
        mapping = Mapping(
            input_combination=[{"type": EV_KEY, "code": KEY_1}],
            target_uinput="keyboard",
            output_symbol="key(a)",
        )
        mapping.set_combination_changed_callback(lambda *args: None)
        self.assertRaisesRegex(
            ValidationError,
            # the error should mention
            # - the symbol
            # - the current incorrect target
            # - the target that works for this symbol
            ".*BTN_A.*keyboard.*gamepad",
            mapping.__setattr__,
            "output_symbol",
            "key(BTN_A)",
        )


@test_setup
class TestUIMapping(unittest.IsolatedAsyncioTestCase):
    def test_init(self):
        """Should be able to initialize without throwing errors."""
        UIMapping()

    def test_is_valid(self):
        """Should be invalid at first and become valid once all data is provided."""
        mapping = UIMapping()
        self.assertFalse(mapping.is_valid())

        mapping.input_combination = [{"type": EV_KEY, "code": KEY_1}]
        mapping.output_symbol = "a"
        self.assertFalse(mapping.is_valid())
        mapping.target_uinput = "keyboard"
        self.assertTrue(mapping.is_valid())

    def test_updates_validation_error(self):
        mapping = UIMapping()
        self.assertGreaterEqual(len(mapping.get_error().errors()), 2)
        mapping.input_combination = [{"type": EV_KEY, "code": KEY_1}]
        mapping.output_symbol = "a"
        self.assertIn(
            "1 validation error for Mapping\ntarget_uinput",
            str(mapping.get_error()),
        )
        mapping.target_uinput = "keyboard"
        self.assertTrue(mapping.is_valid())
        self.assertIsNone(mapping.get_error())

    def test_copy_returns_ui_mapping(self):
        """Copy should also be a UIMapping with all the invalid data."""
        mapping = UIMapping()
        mapping_2 = mapping.copy()
        self.assertIsInstance(mapping_2, UIMapping)
        self.assertEqual(
            mapping_2.input_combination, InputCombination.empty_combination()
        )
        self.assertIsNone(mapping_2.output_symbol)

    def test_get_bus_massage(self):
        mapping = UIMapping()
        mapping_2 = mapping.get_bus_message()
        self.assertEqual(mapping_2.message_type, MessageType.mapping)

        with self.assertRaises(TypeError):
            # the massage should be immutable
            mapping_2.output_symbol = "a"
        self.assertIsNone(mapping_2.output_symbol)

        # the original should be not immutable
        mapping.output_symbol = "a"
        self.assertEqual(mapping.output_symbol, "a")

    def test_has_input_defined(self):
        mapping = UIMapping()
        self.assertFalse(mapping.has_input_defined())
        mapping.input_combination = InputCombination([InputConfig(type=EV_KEY, code=1)])
        self.assertTrue(mapping.has_input_defined())


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

"""Starts the root reader-service."""

import asyncio
import atexit
import os
import signal
import sys
import time
from argparse import ArgumentParser

from inputremapper.bin.process_utils import ProcessUtils
from inputremapper.groups import _Groups
from inputremapper.gui.reader_service import ReaderService
from inputremapper.injection.global_uinputs import GlobalUInputs, FrontendUInput
from inputremapper.logging.logger import logger


class InputRemapperReaderServiceBin:
    @staticmethod
    def main() -> None:
        parser = ArgumentParser()
        parser.add_argument(
            "-d",
            "--debug",
            action="store_true",
            dest="debug",
            help="Displays additional debug information",
            default=False,
        )

        options = parser.parse_args(sys.argv[1:])

        logger.update_verbosity(options.debug)

        if ProcessUtils.count_python_processes("input-remapper-reader-service") >= 2:
            logger.warning(
                "Another input-remapper-reader-service process is already running. "
                "This can cause problems while recording keys"
            )

        if os.getuid() != 0:
            logger.warning("The reader-service usually needs elevated privileges")

        if not ReaderService.pipes_exist():
            logger.info("Waiting for pipes to be created by the GUI")
            while not ReaderService.pipes_exist():
                time.sleep(0.5)
                logger.debug("Waiting...")

        def on_exit():
            """Don't remain idle and alive when the GUI exits via ctrl+c."""
            # makes no sense to me, but after the keyboard interrupt it is still
            # waiting for an event to complete (`S` in `ps ax`), even when using
            # sys.exit
            os.kill(os.getpid(), signal.SIGKILL)

        atexit.register(on_exit)
        groups = _Groups()
        global_uinputs = GlobalUInputs(FrontendUInput)
        reader_service = ReaderService(groups, global_uinputs)
        asyncio.run(reader_service.run())



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


"""All components that control a single preset."""


from __future__ import annotations

from collections import defaultdict
from typing import List, Optional, Dict, Union, Callable, Literal, Set

import cairo
from evdev.ecodes import (
    EV_KEY,
    EV_ABS,
    EV_REL,
    BTN_LEFT,
    BTN_MIDDLE,
    BTN_RIGHT,
    BTN_EXTRA,
    BTN_SIDE,
)
from gi.repository import Gtk, GtkSource, Gdk

from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.configs.keyboard_layout import keyboard_layout, XKB_KEYCODE_OFFSET
from inputremapper.configs.mapping import MappingData, MappingType
from inputremapper.groups import DeviceType
from inputremapper.gui.components.output_type_names import OutputTypeNames
from inputremapper.gui.controller import Controller
from inputremapper.gui.gettext import _
from inputremapper.gui.messages.message_broker import (
    MessageBroker,
    MessageType,
)
from inputremapper.gui.messages.message_data import (
    UInputsData,
    PresetData,
    CombinationUpdate,
)
from inputremapper.gui.utils import HandlerDisabled, Colors
from inputremapper.injection.mapping_handlers.axis_transform import Transformation
from inputremapper.input_event import InputEvent
from inputremapper.utils import get_evdev_constant_name

Capabilities = Dict[int, List]

SET_KEY_FIRST = _("Record the input first")

ICON_NAMES = {
    DeviceType.GAMEPAD: "input-gaming",
    DeviceType.MOUSE: "input-mouse",
    DeviceType.KEYBOARD: "input-keyboard",
    DeviceType.GRAPHICS_TABLET: "input-tablet",
    DeviceType.TOUCHPAD: "input-touchpad",
    DeviceType.UNKNOWN: None,
}

# sort types that most devices would fall in easily to the right.
ICON_PRIORITIES = [
    DeviceType.GRAPHICS_TABLET,
    DeviceType.TOUCHPAD,
    DeviceType.GAMEPAD,
    DeviceType.MOUSE,
    DeviceType.KEYBOARD,
    DeviceType.UNKNOWN,
]


class TargetSelection:
    """The dropdown menu to select the targe_uinput of the active_mapping,

    For example "keyboard" or "gamepad".
    """

    _mapping: Optional[MappingData] = None

    def __init__(
        self,
        message_broker: MessageBroker,
        controller: Controller,
        combobox: Gtk.ComboBox,
    ):
        self._message_broker = message_broker
        self._controller = controller
        self._gui = combobox

        self._message_broker.subscribe(MessageType.uinputs, self._on_uinputs_changed)
        self._message_broker.subscribe(MessageType.mapping, self._on_mapping_loaded)
        self._gui.connect("changed", self._on_gtk_target_selected)

    def _select_current_target(self):
        """Select the currently configured target."""
        if self._mapping is not None:
            with HandlerDisabled(self._gui, self._on_gtk_target_selected):
                self._gui.set_active_id(self._mapping.target_uinput)

    def _on_uinputs_changed(self, data: UInputsData):
        target_store = Gtk.ListStore(str)
        for uinput in data.uinputs.keys():
            target_store.append([uinput])

        self._gui.set_model(target_store)
        renderer_text = Gtk.CellRendererText()
        self._gui.pack_start(renderer_text, False)
        self._gui.add_attribute(renderer_text, "text", 0)
        self._gui.set_id_column(0)

        self._select_current_target()

    def _on_mapping_loaded(self, mapping: MappingData):
        self._mapping = mapping
        self._select_current_target()

    def _on_gtk_target_selected(self, *_):
        target = self._gui.get_active_id()
        self._controller.update_mapping(target_uinput=target)


class MappingListBox:
    """The listbox showing all available mapping in the active_preset."""

    def __init__(
        self,
        message_broker: MessageBroker,
        controller: Controller,
        listbox: Gtk.ListBox,
    ):
        self._message_broker = message_broker
        self._controller = controller
        self._gui = listbox
        self._gui.set_sort_func(self._sort_func)

        self._message_broker.subscribe(MessageType.preset, self._on_preset_changed)
        self._message_broker.subscribe(MessageType.mapping, self._on_mapping_changed)
        self._gui.connect("row-selected", self._on_gtk_mapping_selected)

    @staticmethod
    def _sort_func(row1: MappingSelectionLabel, row2: MappingSelectionLabel) -> int:
        """Sort alphanumerical by name."""
        if row1.combination == InputCombination.empty_combination():
            return 1
        if row2.combination == InputCombination.empty_combination():
            return 0

        return 0 if row1.name < row2.name else 1

    def _on_preset_changed(self, data: PresetData):
        selection_labels = self._gui.get_children()
        for selection_label in selection_labels:
            selection_label.cleanup()
            self._gui.remove(selection_label)

        if not data.mappings:
            return

        for mapping in data.mappings:
            selection_label = MappingSelectionLabel(
                self._message_broker,
                self._controller,
                mapping.format_name(),
                mapping.input_combination,
            )
            self._gui.insert(selection_label, -1)
        self._gui.invalidate_sort()

    def _on_mapping_changed(self, mapping: MappingData):
        with HandlerDisabled(self._gui, self._on_gtk_mapping_selected):
            combination = mapping.input_combination

            for row in self._gui.get_children():
                if row.combination == combination:
                    self._gui.select_row(row)

    def _on_gtk_mapping_selected(self, _, row: Optional[MappingSelectionLabel]):
        if not row:
            return
        self._controller.load_mapping(row.combination)


class MappingSelectionLabel(Gtk.ListBoxRow):
    """The ListBoxRow representing a mapping inside the MappingListBox."""

    __gtype_name__ = "MappingSelectionLabel"

    def __init__(
        self,
        message_broker: MessageBroker,
        controller: Controller,
        name: Optional[str],
        combination: InputCombination,
    ):
        super().__init__()
        self._message_broker = message_broker
        self._controller = controller

        if not name:
            name = combination.beautify()

        self.name = name
        self.combination = combination
        # add hotkey handler
        self.connect("key-press-event", self._on_key_press)

        # Make the child label widget break lines, important for
        # long combinations
        self.label = Gtk.Label()
        self.label.set_line_wrap(True)
        self.label.set_line_wrap_mode(Gtk.WrapMode.WORD)
        self.label.set_justify(Gtk.Justification.CENTER)
        # set the name or combination.beautify as label
        self.label.set_label(self.name)

        self.label.set_margin_top(11)
        self.label.set_margin_bottom(11)

        # button to edit the name of the mapping
        self.edit_btn = Gtk.Button()
        self.edit_btn.set_relief(Gtk.ReliefStyle.NONE)
        self.edit_btn.set_image(
            Gtk.Image.new_from_icon_name(Gtk.STOCK_EDIT, Gtk.IconSize.MENU)
        )
        self.edit_btn.set_tooltip_text(_("Change Mapping Name") + " (F2)")
        self.edit_btn.set_margin_top(4)
        self.edit_btn.set_margin_bottom(4)
        self.edit_btn.connect("clicked", self._set_edit_mode)

        self.name_input = Gtk.Entry()
        self.name_input.set_text(self.name)
        self.name_input.set_halign(Gtk.Align.FILL)
        self.name_input.set_margin_top(4)
        self.name_input.set_margin_bottom(4)
        self.name_input.connect("activate", self._on_gtk_rename_finished)
        self.name_input.connect("key-press-event", self._on_gtk_rename_abort)

        self._box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        self._box.set_center_widget(self.label)
        self._box.add(self.edit_btn)
        self._box.set_child_packing(self.edit_btn, False, False, 4, Gtk.PackType.END)
        self._box.add(self.name_input)
        self._box.set_child_packing(self.name_input, True, True, 4, Gtk.PackType.START)

        self.add(self._box)
        self.show_all()
        self._message_broker.subscribe(MessageType.mapping, self._on_mapping_changed)
        self._message_broker.subscribe(
            MessageType.combination_update, self._on_combination_update
        )

        self.edit_btn.hide()
        self.name_input.hide()

    def __repr__(self):
        return f"<MappingSelectionLabel for {self.combination} as {self.name} at {hex(id(self))}>"

    def _set_not_selected(self):
        self.edit_btn.hide()
        self.name_input.hide()
        self.label.show()

    def _set_selected(self):
        self.label.set_label(self.name)
        self.edit_btn.show()
        self.name_input.hide()
        self.label.show()

    def _set_edit_mode(self, *_):
        self.name_input.set_text(self.name)
        self.label.hide()
        self.name_input.show()
        self._controller.set_focus(self.name_input)

    def _on_key_press(self, widget, event):
        """ "Mapping Selection Row Label key handler for hotkeys"""
        if self.label.is_visible():
            # hotkeys are only meaningful when not already in edit-mapping-name-mode
            if event.keyval == Gdk.KEY_F2:
                self._set_edit_mode()
            elif event.keyval == Gdk.KEY_Delete:
                self._controller.delete_mapping()

    def _on_mapping_changed(self, mapping: MappingData):
        if mapping.input_combination != self.combination:
            self._set_not_selected()
            return
        self.name = mapping.format_name()
        self._set_selected()
        self.get_parent().invalidate_sort()

    def _on_combination_update(self, data: CombinationUpdate):
        if data.old_combination == self.combination and self.is_selected():
            self.combination = data.new_combination

    def _on_gtk_rename_finished(self, *_):
        name = self.name_input.get_text()
        if name.lower().strip() == self.combination.beautify().lower():
            name = ""
        self.name = name
        self._set_selected()
        self._controller.update_mapping(name=name)

    def _on_gtk_rename_abort(self, _, key_event: Gdk.EventKey):
        if key_event.keyval == Gdk.KEY_Escape:
            self._set_selected()

    def cleanup(self) -> None:
        """Clean up message listeners. Execute before removing from gui!"""
        self._message_broker.unsubscribe(self._on_mapping_changed)
        self._message_broker.unsubscribe(self._on_combination_update)


class GdkEventRecorder:
    """Records events delivered by GDK, similar to the ReaderService/ReaderClient."""

    _combination: List[int]
    _pressed: Set[int]

    __gtype_name__ = "GdkEventRecorder"

    def __init__(self, window: Gtk.Window, gui: Gtk.Label):
        super().__init__()
        self._combination = []
        self._pressed = set()
        self._gui = gui
        window.connect("event", self._on_gtk_event)

    def _get_button_code(self, event: Gdk.Event):
        """Get the evdev code for the given event."""
        return {
            Gdk.BUTTON_MIDDLE: BTN_MIDDLE,
            Gdk.BUTTON_PRIMARY: BTN_LEFT,
            Gdk.BUTTON_SECONDARY: BTN_RIGHT,
            9: BTN_EXTRA,
            8: BTN_SIDE,
        }.get(event.get_button().button)

    def _reset(self, event: Gdk.Event):
        """If a new combination is being typed, start from scratch."""
        gdk_event_type: int = event.type

        is_press = gdk_event_type in [
            Gdk.EventType.KEY_PRESS,
            Gdk.EventType.BUTTON_PRESS,
        ]

        if len(self._pressed) == 0 and is_press:
            self._combination = []

    def _press(self, event: Gdk.Event):
        """Remember pressed keys, write down combinations."""
        gdk_event_type: int = event.type

        if gdk_event_type == Gdk.EventType.KEY_PRESS:
            code = event.hardware_keycode - XKB_KEYCODE_OFFSET
            if code not in self._combination:
                self._combination.append(code)

            self._pressed.add(code)

        if gdk_event_type == Gdk.EventType.BUTTON_PRESS:
            code = self._get_button_code(event)
            if code not in self._combination:
                self._combination.append(code)

            self._pressed.add(code)

    def _release(self, event: Gdk.Event):
        """Clear pressed keys if this is a release event."""
        if event.type in [Gdk.EventType.KEY_RELEASE, Gdk.EventType.BUTTON_RELEASE]:
            self._pressed = set()

    def _display(self, event):
        """Show the recorded combination in the gui."""
        is_press = event.type in [
            Gdk.EventType.KEY_PRESS,
            Gdk.EventType.BUTTON_PRESS,
        ]

        if is_press and len(self._combination) > 0:
            names = [
                keyboard_layout.get_name(code)
                for code in self._combination
                if code is not None and keyboard_layout.get_name(code) is not None
            ]
            self._gui.set_text(" + ".join(names))

    def _on_gtk_event(self, _, event: Gdk.Event):
        """For all sorts of input events that gtk cares about."""
        self._reset(event)
        self._release(event)
        self._press(event)
        self._display(event)


class CodeEditor:
    """The editor used to edit the output_symbol of the active_mapping."""

    placeholder: str = _("Enter your output here")

    def __init__(
        self,
        message_broker: MessageBroker,
        controller: Controller,
        editor: GtkSource.View,
    ):
        self._message_broker = message_broker
        self._controller = controller
        self.gui = editor

        # without this the wrapping ScrolledWindow acts weird when new lines are added,
        # not offering enough space to the text editor so the whole thing is suddenly
        # scrollable by a few pixels.
        # Found this after making blind guesses with settings in glade, and then
        # actually looking at the snapshot preview! In glades editor this didn't have an
        # effect.
        self.gui.set_resize_mode(Gtk.ResizeMode.IMMEDIATE)

        # Syntax Highlighting
        # TODO there are some similarities with python, but overall it's quite useless.
        #  commented out until there is proper highlighting for input-remappers syntax.
        # Thanks to https://github.com/wolfthefallen/py-GtkSourceCompletion-example
        # language_manager = GtkSource.LanguageManager()
        # fun fact: without saving LanguageManager into its own variable it doesn't work
        #  python = language_manager.get_language("python")
        # source_view.get_buffer().set_language(python)

        self._update_placeholder()

        self.gui.get_buffer().connect("changed", self._on_gtk_changed)
        self.gui.connect("focus-in-event", self._update_placeholder)
        self.gui.connect("focus-out-event", self._update_placeholder)
        self._connect_message_listener()

    def _update_placeholder(self, *_):
        buffer = self.gui.get_buffer()
        code = buffer.get_text(buffer.get_start_iter(), buffer.get_end_iter(), True)

        # test for incorrect states and fix them, without causing side effects
        with HandlerDisabled(buffer, self._on_gtk_changed):
            if self.gui.has_focus() and code == self.placeholder:
                # hide the placeholder
                buffer.set_text("")
                self.gui.get_style_context().remove_class("opaque-text")
            elif code == "":
                # show the placeholder instead
                buffer.set_text(self.placeholder)
                self.gui.get_style_context().add_class("opaque-text")
            elif code != "":
                # something is written, ensure the opacity is correct
                self.gui.get_style_context().remove_class("opaque-text")

    def _shows_placeholder(self):
        buffer = self.gui.get_buffer()
        code = buffer.get_text(buffer.get_start_iter(), buffer.get_end_iter(), True)
        return code == self.placeholder

    @property
    def code(self) -> str:
        """Get the user-defined macro code string."""
        if self._shows_placeholder():
            return ""

        buffer = self.gui.get_buffer()
        return buffer.get_text(buffer.get_start_iter(), buffer.get_end_iter(), True)

    @code.setter
    def code(self, code: str) -> None:
        """Set the text without triggering any events."""
        buffer = self.gui.get_buffer()
        with HandlerDisabled(buffer, self._on_gtk_changed):
            buffer.set_text(code)
            self._update_placeholder()
            self.gui.do_move_cursor(self.gui, Gtk.MovementStep.BUFFER_ENDS, -1, False)

    def _connect_message_listener(self):
        self._message_broker.subscribe(
            MessageType.mapping,
            self._on_mapping_loaded,
        )
        self._message_broker.subscribe(
            MessageType.recording_finished,
            self._on_recording_finished,
        )

    def _toggle_line_numbers(self):
        """Show line numbers if multiline, otherwise remove them."""
        if "\n" in self.code:
            self.gui.set_show_line_numbers(True)
            # adds a bit of space between numbers and text:
            self.gui.set_show_line_marks(True)
            self.gui.set_monospace(True)
            self.gui.get_style_context().add_class("multiline")
        else:
            self.gui.set_show_line_numbers(False)
            self.gui.set_show_line_marks(False)
            self.gui.set_monospace(False)
            self.gui.get_style_context().remove_class("multiline")

    def _on_gtk_changed(self, *_):
        if self._shows_placeholder():
            return

        self._controller.update_mapping(output_symbol=self.code)

    def _on_mapping_loaded(self, mapping: MappingData):
        code = SET_KEY_FIRST
        if not self._controller.is_empty_mapping():
            code = mapping.output_symbol or ""

        if self.code.strip().lower() != code.strip().lower():
            self.code = code

        self._toggle_line_numbers()

    def _on_recording_finished(self, _):
        self._controller.set_focus(self.gui)


class RequireActiveMapping:
    """Disable the widget if no mapping is selected."""

    def __init__(
        self,
        message_broker: MessageBroker,
        widget: Gtk.ToggleButton,
        require_recorded_input: bool,
    ):
        self._widget = widget
        self._default_tooltip = self._widget.get_tooltip_text()
        self._require_recorded_input = require_recorded_input

        self._active_preset: Optional[PresetData] = None
        self._active_mapping: Optional[MappingData] = None

        message_broker.subscribe(MessageType.preset, self._on_preset)
        message_broker.subscribe(MessageType.mapping, self._on_mapping)

    def _on_preset(self, preset_data: PresetData):
        self._active_preset = preset_data
        self._check()

    def _on_mapping(self, mapping_data: MappingData):
        self._active_mapping = mapping_data
        self._check()

    def _check(self, *__):
        if not self._active_preset or len(self._active_preset.mappings) == 0:
            self._disable()
            self._widget.set_tooltip_text(_("Add a mapping first"))
            return

        if (
            self._require_recorded_input
            and self._active_mapping
            and not self._active_mapping.has_input_defined()
        ):
            self._disable()
            self._widget.set_tooltip_text(_("Record input first"))
            return

        self._enable()
        self._widget.set_tooltip_text(self._default_tooltip)

    def _enable(self):
        self._widget.set_sensitive(True)
        self._widget.set_opacity(1)

    def _disable(self):
        self._widget.set_sensitive(False)
        self._widget.set_opacity(0.5)


class RecordingToggle:
    """The toggle that starts input recording for the active_mapping."""

    def __init__(
        self,
        message_broker: MessageBroker,
        controller: Controller,
        toggle: Gtk.ToggleButton,
    ):
        self._message_broker = message_broker
        self._controller = controller
        self._gui = toggle

        toggle.connect("toggled", self._on_gtk_toggle)
        # Don't leave the input when using arrow keys or tab. wait for the
        # window to consume the keycode from the reader. I.e. a tab input should
        # be recorded, instead of causing the recording to stop.
        toggle.connect("key-press-event", lambda *args: Gdk.EVENT_STOP)
        self._message_broker.subscribe(
            MessageType.recording_finished,
            self._on_recording_finished,
        )

        RequireActiveMapping(
            message_broker,
            toggle,
            require_recorded_input=False,
        )

    def _on_gtk_toggle(self, *__):
        if self._gui.get_active():
            self._controller.start_key_recording()
        else:
            self._controller.stop_key_recording()

    def _on_recording_finished(self, __):
        with HandlerDisabled(self._gui, self._on_gtk_toggle):
            self._gui.set_active(False)


class RecordingStatus:
    """Displays if keys are being recorded for a mapping."""

    def __init__(
        self,
        message_broker: MessageBroker,
        label: Gtk.Label,
    ):
        self._gui = label

        message_broker.subscribe(
            MessageType.recording_started,
            self._on_recording_started,
        )

        message_broker.subscribe(
            MessageType.recording_finished,
            self._on_recording_finished,
        )

    def _on_recording_started(self, _):
        self._gui.set_visible(True)

    def _on_recording_finished(self, _):
        self._gui.set_visible(False)


class AutoloadSwitch:
    """The switch used to toggle the autoload state of the active_preset."""

    def __init__(
        self, message_broker: MessageBroker, controller: Controller, switch: Gtk.Switch
    ):
        self._message_broker = message_broker
        self._controller = controller
        self._gui = switch

        self._gui.connect("state-set", self._on_gtk_toggle)
        self._message_broker.subscribe(MessageType.preset, self._on_preset_changed)

    def _on_preset_changed(self, data: PresetData):
        with HandlerDisabled(self._gui, self._on_gtk_toggle):
            self._gui.set_active(data.autoload)

    def _on_gtk_toggle(self, *_):
        self._controller.set_autoload(self._gui.get_active())


class ReleaseCombinationSwitch:
    """The switch used to set the active_mapping.release_combination_keys parameter."""

    def __init__(
        self, message_broker: MessageBroker, controller: Controller, switch: Gtk.Switch
    ):
        self._message_broker = message_broker
        self._controller = controller
        self._gui = switch

        self._gui.connect("state-set", self._on_gtk_toggle)
        self._message_broker.subscribe(MessageType.mapping, self._on_mapping_changed)

    def _on_mapping_changed(self, data: MappingData):
        with HandlerDisabled(self._gui, self._on_gtk_toggle):
            self._gui.set_active(data.release_combination_keys)

    def _on_gtk_toggle(self, *_):
        self._controller.update_mapping(release_combination_keys=self._gui.get_active())


class InputConfigEntry(Gtk.ListBoxRow):
    """The ListBoxRow representing a single input config inside the CombinationListBox."""

    __gtype_name__ = "InputConfigEntry"

    def __init__(self, event: InputConfig, controller: Controller):
        super().__init__()

        self.input_event = event
        self._controller = controller

        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=4)
        hbox.set_margin_start(12)

        label = Gtk.Label()
        label.set_label(event.description())
        hbox.pack_start(label, False, False, 0)

        up_btn = Gtk.Button()
        up_btn.set_halign(Gtk.Align.END)
        up_btn.set_relief(Gtk.ReliefStyle.NONE)
        up_btn.get_style_context().add_class("no-v-padding")
        up_img = Gtk.Image.new_from_icon_name("go-up", Gtk.IconSize.BUTTON)
        up_btn.add(up_img)

        down_btn = Gtk.Button()
        down_btn.set_halign(Gtk.Align.END)
        down_btn.set_relief(Gtk.ReliefStyle.NONE)
        down_btn.get_style_context().add_class("no-v-padding")
        down_img = Gtk.Image.new_from_icon_name("go-down", Gtk.IconSize.BUTTON)
        down_btn.add(down_img)

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        vbox.pack_start(up_btn, False, True, 0)
        vbox.pack_end(down_btn, False, True, 0)
        hbox.pack_end(vbox, False, False, 0)

        up_btn.connect(
            "clicked",
            lambda *_: self._controller.move_input_config_in_combination(
                self.input_event, "up"
            ),
        )
        down_btn.connect(
            "clicked",
            lambda *_: self._controller.move_input_config_in_combination(
                self.input_event, "down"
            ),
        )
        self.add(hbox)
        self.show_all()

        # only used in testing
        self._up_btn = up_btn
        self._down_btn = down_btn


class CombinationListbox:
    """The ListBox with all the events inside active_mapping.input_combination."""

    def __init__(
        self,
        message_broker: MessageBroker,
        controller: Controller,
        listbox: Gtk.ListBox,
    ):
        self._message_broker = message_broker
        self._controller = controller
        self._gui = listbox
        self._combination: Optional[InputCombination] = None

        self._message_broker.subscribe(
            MessageType.mapping,
            self._on_mapping_changed,
        )
        self._message_broker.subscribe(
            MessageType.selected_event,
            self._on_event_changed,
        )

        self._gui.connect("row-selected", self._on_gtk_row_selected)

    def _select_row(self, event: InputEvent):
        for row in self._gui.get_children():
            if row.input_event == event:
                self._gui.select_row(row)

    def _on_mapping_changed(self, mapping: MappingData):
        if self._combination == mapping.input_combination:
            return

        event_entries = self._gui.get_children()
        for event_entry in event_entries:
            self._gui.remove(event_entry)

        if self._controller.is_empty_mapping():
            self._combination = None
        else:
            self._combination = mapping.input_combination
            for event in self._combination:
                self._gui.insert(InputConfigEntry(event, self._controller), -1)

    def _on_event_changed(self, event: InputEvent):
        with HandlerDisabled(self._gui, self._on_gtk_row_selected):
            self._select_row(event)

    def _on_gtk_row_selected(self, *_):
        for row in self._gui.get_children():
            if row.is_selected():
                self._controller.load_input_config(row.input_event)
                break


class AnalogInputSwitch:
    """The switch that marks the active_input_config as analog input."""

    def __init__(
        self,
        message_broker: MessageBroker,
        controller: Controller,
        gui: Gtk.Switch,
    ):
        self._message_broker = message_broker
        self._controller = controller
        self._gui = gui
        self._input_config: Optional[InputConfig] = None

        self._gui.connect("state-set", self._on_gtk_toggle)
        self._message_broker.subscribe(MessageType.selected_event, self._on_event)

    def _on_event(self, input_cfg: InputConfig):
        with HandlerDisabled(self._gui, self._on_gtk_toggle):
            self._gui.set_active(input_cfg.defines_analog_input)
            self._input_config = input_cfg

        if input_cfg.type == EV_KEY:
            self._gui.set_sensitive(False)
            self._gui.set_opacity(0.5)
        else:
            self._gui.set_sensitive(True)
            self._gui.set_opacity(1)

    def _on_gtk_toggle(self, *_):
        self._controller.set_event_as_analog(self._gui.get_active())


class TriggerThresholdInput:
    """The number selection used to set the speed or position threshold of the
    active_input_config when it is an ABS or REL event used as a key."""

    def __init__(
        self,
        message_broker: MessageBroker,
        controller: Controller,
        gui: Gtk.SpinButton,
    ):
        self._message_broker = message_broker
        self._controller = controller
        self._gui = gui
        self._input_config: Optional[InputConfig] = None

        self._gui.set_increments(1, 1)
        self._gui.connect("value-changed", self._on_gtk_changed)
        self._message_broker.subscribe(MessageType.selected_event, self._on_event)

    def _on_event(self, input_config: InputConfig):
        if input_config.type == EV_KEY:
            self._gui.set_sensitive(False)
            self._gui.set_opacity(0.5)
        elif input_config.type == EV_ABS:
            self._gui.set_sensitive(True)
            self._gui.set_opacity(1)
            self._gui.set_range(-99, 99)
        else:
            self._gui.set_sensitive(True)
            self._gui.set_opacity(1)
            self._gui.set_range(-999, 999)

        with HandlerDisabled(self._gui, self._on_gtk_changed):
            self._gui.set_value(input_config.analog_threshold or 0)
            self._input_config = input_config

    def _on_gtk_changed(self, *_):
        self._controller.update_input_config(
            self._input_config.modify(analog_threshold=int(self._gui.get_value()))
        )


class ReleaseTimeoutInput:
    """The number selector used to set the active_mapping.release_timeout parameter."""

    def __init__(
        self,
        message_broker: MessageBroker,
        controller: Controller,
        gui: Gtk.SpinButton,
    ):
        self._message_broker = message_broker
        self._controller = controller
        self._gui = gui

        self._gui.set_increments(0.01, 0.01)
        self._gui.set_range(0, 2)
        self._gui.connect("value-changed", self._on_gtk_changed)
        self._message_broker.subscribe(MessageType.mapping, self._on_mapping_message)

    def _on_mapping_message(self, mapping: MappingData):
        if EV_REL in [event.type for event in mapping.input_combination]:
            self._gui.set_sensitive(True)
            self._gui.set_opacity(1)
        else:
            self._gui.set_sensitive(False)
            self._gui.set_opacity(0.5)

        with HandlerDisabled(self._gui, self._on_gtk_changed):
            self._gui.set_value(mapping.release_timeout)

    def _on_gtk_changed(self, *_):
        self._controller.update_mapping(release_timeout=self._gui.get_value())


class RelativeInputCutoffInput:
    """The number selector to set active_mapping.rel_to_abs_input_cutoff."""

    def __init__(
        self,
        message_broker: MessageBroker,
        controller: Controller,
        gui: Gtk.SpinButton,
    ):
        self._message_broker = message_broker
        self._controller = controller
        self._gui = gui

        self._gui.set_increments(1, 1)
        self._gui.set_range(1, 1000)
        self._gui.connect("value-changed", self._on_gtk_changed)
        self._message_broker.subscribe(MessageType.mapping, self._on_mapping_message)

    def _on_mapping_message(self, mapping: MappingData):
        if (
            EV_REL in [event.type for event in mapping.input_combination]
            and mapping.output_type == EV_ABS
        ):
            self._gui.set_sensitive(True)
            self._gui.set_opacity(1)
        else:
            self._gui.set_sensitive(False)
            self._gui.set_opacity(0.5)

        with HandlerDisabled(self._gui, self._on_gtk_changed):
            self._gui.set_value(mapping.rel_to_abs_input_cutoff)

    def _on_gtk_changed(self, *_):
        self._controller.update_mapping(rel_xy_cutoff=self._gui.get_value())


class OutputAxisSelector:
    """The dropdown menu used to select the output axis if the active_mapping is a
    mapping targeting an analog axis

    modifies the active_mapping.output_code and active_mapping.output_type parameters
    """

    def __init__(
        self,
        message_broker: MessageBroker,
        controller: Controller,
        gui: Gtk.ComboBox,
    ):
        self._message_broker = message_broker
        self._controller = controller
        self._gui = gui
        self._uinputs: Dict[str, Capabilities] = {}
        self.model = Gtk.ListStore(str, str)

        self._current_target: Optional[str] = None

        self._gui.set_model(self.model)
        renderer_text = Gtk.CellRendererText()
        self._gui.pack_start(renderer_text, False)
        self._gui.add_attribute(renderer_text, "text", 1)
        self._gui.set_id_column(0)

        self._gui.connect("changed", self._on_gtk_select_axis)
        self._message_broker.subscribe(MessageType.mapping, self._on_mapping_message)
        self._message_broker.subscribe(MessageType.uinputs, self._on_uinputs_message)

    def _set_model(self, target: Optional[str]):
        if target == self._current_target:
            return

        self.model.clear()
        self.model.append(["None, None", _("No Axis")])

        if target is not None:
            capabilities = self._uinputs.get(target) or defaultdict(list)
            types_codes = [
                (EV_ABS, code) for code, absinfo in capabilities.get(EV_ABS) or ()
            ]
            types_codes.extend(
                (EV_REL, code) for code in capabilities.get(EV_REL) or ()
            )
            for type_, code in types_codes:
                key_name = get_evdev_constant_name(type_, code)
                if isinstance(key_name, list):
                    key_name = key_name[0]
                self.model.append([f"{type_}, {code}", key_name])

        self._current_target = target

    def _on_mapping_message(self, mapping: MappingData):
        with HandlerDisabled(self._gui, self._on_gtk_select_axis):
            self._set_model(mapping.target_uinput)
            self._gui.set_active_id(f"{mapping.output_type}, {mapping.output_code}")

    def _on_uinputs_message(self, uinputs: UInputsData):
        self._uinputs = uinputs.uinputs

    def _on_gtk_select_axis(self, *_):
        if self._gui.get_active_id() == "None, None":
            type_code = (None, None)
        else:
            type_code = tuple(int(i) for i in self._gui.get_active_id().split(","))
        self._controller.update_mapping(
            output_type=type_code[0], output_code=type_code[1]
        )


class KeyAxisStackSwitcher:
    """The controls used to switch between the gui to modify a key-mapping or
    an analog-axis mapping."""

    def __init__(
        self,
        message_broker: MessageBroker,
        controller: Controller,
        stack: Gtk.Stack,
        key_macro_toggle: Gtk.ToggleButton,
        analog_toggle: Gtk.ToggleButton,
    ):
        self._message_broker = message_broker
        self._controller = controller
        self._stack = stack
        self._key_macro_toggle = key_macro_toggle
        self._analog_toggle = analog_toggle

        self._key_macro_toggle.connect("toggled", self._on_gtk_toggle)
        self._analog_toggle.connect("toggled", self._on_gtk_toggle)
        self._message_broker.subscribe(MessageType.mapping, self._on_mapping_message)

    def _set_active(self, mapping_type: Literal["key_macro", "analog"]):
        if mapping_type == MappingType.ANALOG.value:
            self._stack.set_visible_child_name(OutputTypeNames.analog_axis)
            active = self._analog_toggle
            inactive = self._key_macro_toggle
        else:
            self._stack.set_visible_child_name(OutputTypeNames.key_or_macro)
            active = self._key_macro_toggle
            inactive = self._analog_toggle

        with HandlerDisabled(active, self._on_gtk_toggle):
            active.set_active(True)
        with HandlerDisabled(inactive, self._on_gtk_toggle):
            inactive.set_active(False)

    def _on_mapping_message(self, mapping: MappingData):
        # fist check the actual mapping
        if mapping.mapping_type == MappingType.ANALOG.value:
            self._set_active(MappingType.ANALOG.value)

        if mapping.mapping_type == MappingType.KEY_MACRO.value:
            self._set_active(MappingType.KEY_MACRO.value)

    def _on_gtk_toggle(self, btn: Gtk.ToggleButton):
        # get_active returns the new toggle state already
        was_active = not btn.get_active()

        if was_active:
            # cannot deactivate manually
            with HandlerDisabled(btn, self._on_gtk_toggle):
                btn.set_active(True)
            return

        if btn is self._key_macro_toggle:
            self._controller.update_mapping(mapping_type=MappingType.KEY_MACRO.value)
        else:
            self._controller.update_mapping(mapping_type=MappingType.ANALOG.value)


class TransformationDrawArea:
    """The graph which shows the relation between input- and output-axis."""

    def __init__(
        self,
        message_broker: MessageBroker,
        controller: Controller,
        gui: Gtk.DrawingArea,
    ):
        self._message_broker = message_broker
        self._controller = controller
        self._gui = gui

        self._transformation: Callable[[Union[float, int]], float] = lambda x: x

        self._gui.connect("draw", self._on_gtk_draw)
        self._message_broker.subscribe(MessageType.mapping, self._on_mapping_message)

    def _on_mapping_message(self, mapping: MappingData):
        self._transformation = Transformation(
            100, -100, mapping.deadzone, mapping.gain, mapping.expo
        )
        self._gui.queue_draw()

    def _on_gtk_draw(self, _, context: cairo.Context):
        points = [
            (x / 200 + 0.5, -0.5 * self._transformation(x) + 0.5)
            # leave some space left and right for the lineCap to be visible
            for x in range(-97, 97)
        ]
        width = self._gui.get_allocated_width()
        height = self._gui.get_allocated_height()
        b = min((width, height))
        scaled_points = [(x * b, y * b) for x, y in points]

        # x arrow
        context.move_to(0 * b, 0.5 * b)
        context.line_to(1 * b, 0.5 * b)
        context.line_to(0.96 * b, 0.52 * b)
        context.move_to(1 * b, 0.5 * b)
        context.line_to(0.96 * b, 0.48 * b)

        # y arrow
        context.move_to(0.5 * b, 1 * b)
        context.line_to(0.5 * b, 0)
        context.line_to(0.48 * b, 0.04 * b)
        context.move_to(0.5 * b, 0)
        context.line_to(0.52 * b, 0.04 * b)

        context.set_line_width(2)
        arrow_color = Gdk.RGBA(0.5, 0.5, 0.5, 0.2)
        context.set_source_rgba(
            arrow_color.red,
            arrow_color.green,
            arrow_color.blue,
            arrow_color.alpha,
        )
        context.stroke()

        # graph
        context.move_to(*scaled_points[0])
        for scaled_point in scaled_points[1:]:
            # Ploting point
            context.line_to(*scaled_point)

        line_color = Colors.get_accent_color()
        context.set_line_width(3)
        context.set_line_cap(cairo.LineCap.ROUND)
        # the default gtk adwaita highlight color:
        context.set_source_rgba(
            line_color.red,
            line_color.green,
            line_color.blue,
            line_color.alpha,
        )
        context.stroke()


class Sliders:
    """The different sliders to modify the gain, deadzone and expo parameters of the
    active_mapping."""

    def __init__(
        self,
        message_broker: MessageBroker,
        controller: Controller,
        gain: Gtk.Range,
        deadzone: Gtk.Range,
        expo: Gtk.Range,
    ):
        self._message_broker = message_broker
        self._controller = controller
        self._gain = gain
        self._deadzone = deadzone
        self._expo = expo

        self._gain.set_range(-2, 2)
        self._deadzone.set_range(0, 0.9)
        self._expo.set_range(-1, 1)

        self._gain.connect("value-changed", self._on_gtk_gain_changed)
        self._expo.connect("value-changed", self._on_gtk_expo_changed)
        self._deadzone.connect("value-changed", self._on_gtk_deadzone_changed)
        self._message_broker.subscribe(MessageType.mapping, self._on_mapping_message)

    def _on_mapping_message(self, mapping: MappingData):
        with HandlerDisabled(self._gain, self._on_gtk_gain_changed):
            self._gain.set_value(mapping.gain)

        with HandlerDisabled(self._expo, self._on_gtk_expo_changed):
            self._expo.set_value(mapping.expo)

        with HandlerDisabled(self._deadzone, self._on_gtk_deadzone_changed):
            self._deadzone.set_value(mapping.deadzone)

    def _on_gtk_gain_changed(self, *_):
        self._controller.update_mapping(gain=self._gain.get_value())

    def _on_gtk_deadzone_changed(self, *_):
        self._controller.update_mapping(deadzone=self._deadzone.get_value())

    def _on_gtk_expo_changed(self, *_):
        self._controller.update_mapping(expo=self._expo.get_value())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

"""Build language files and copy them to the target."""

import glob
import os
import subprocess
from os.path import basename, splitext, join, dirname


def make_lang(root: str) -> None:
    """Build po files as mo files into the expected directory."""
    os.makedirs("mo", exist_ok=True)
    for po_file in glob.glob("po/*.po"):
        lang = splitext(basename(po_file))[0]
        target = join(
            root,
            "usr",
            "share",
            "input-remapper",
            "lang",
            lang,
            "LC_MESSAGES",
            "input-remapper.mo",
        )
        os.makedirs(dirname(target), exist_ok=True)
        print(f"Generating translation {target}")
        subprocess.run(
            [
                "msgfmt",
                "-o",
                target,
                str(po_file),
            ],
            check=True,
        )



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


"""Parse macro code"""

from __future__ import annotations

import re
from typing import Optional, Any, Type, TYPE_CHECKING, Dict, List

from inputremapper.configs.validation_errors import MacroError
from inputremapper.injection.macros.macro import Macro
from inputremapper.injection.macros.raw_value import RawValue
from inputremapper.injection.macros.task import Task
from inputremapper.injection.macros.tasks.add import AddTask
from inputremapper.injection.macros.tasks.event import EventTask
from inputremapper.injection.macros.tasks.hold import HoldTask
from inputremapper.injection.macros.tasks.hold_keys import HoldKeysTask
from inputremapper.injection.macros.tasks.if_eq import IfEqTask
from inputremapper.injection.macros.tasks.if_led import IfNumlockTask, IfCapslockTask
from inputremapper.injection.macros.tasks.if_single import IfSingleTask
from inputremapper.injection.macros.tasks.if_tap import IfTapTask
from inputremapper.injection.macros.tasks.ifeq import DeprecatedIfEqTask
from inputremapper.injection.macros.tasks.key import KeyTask
from inputremapper.injection.macros.tasks.key_down import KeyDownTask
from inputremapper.injection.macros.tasks.key_up import KeyUpTask
from inputremapper.injection.macros.tasks.mod_tap import ModTapTask
from inputremapper.injection.macros.tasks.modify import ModifyTask
from inputremapper.injection.macros.tasks.mouse import MouseTask
from inputremapper.injection.macros.tasks.parallel import ParallelTask
from inputremapper.injection.macros.tasks.mouse_xy import MouseXYTask
from inputremapper.injection.macros.tasks.repeat import RepeatTask
from inputremapper.injection.macros.tasks.set import SetTask
from inputremapper.injection.macros.tasks.toggle import ToggleTask
from inputremapper.injection.macros.tasks.wait import WaitTask
from inputremapper.injection.macros.tasks.wheel import WheelTask
from inputremapper.logging.logger import logger

if TYPE_CHECKING:
    from inputremapper.injection.context import Context
    from inputremapper.configs.mapping import Mapping


class Parser:
    TASK_CLASSES: dict[str, type[Task]] = {
        "modify": ModifyTask,
        "repeat": RepeatTask,
        "toggle": ToggleTask,
        "key": KeyTask,
        "key_down": KeyDownTask,
        "key_up": KeyUpTask,
        "event": EventTask,
        "wait": WaitTask,
        "hold": HoldTask,
        "hold_keys": HoldKeysTask,
        "mouse": MouseTask,
        "mouse_xy": MouseXYTask,
        "wheel": WheelTask,
        "if_eq": IfEqTask,
        "if_numlock": IfNumlockTask,
        "if_capslock": IfCapslockTask,
        "set": SetTask,
        "if_tap": IfTapTask,
        "if_single": IfSingleTask,
        "add": AddTask,
        "mod_tap": ModTapTask,
        "parallel": ParallelTask,
        # Those are only kept for backwards compatibility with old macros. The space for
        # writing macro was very constrained in the past, so shorthands were introduced:
        "m": ModifyTask,
        "r": RepeatTask,
        "k": KeyTask,
        "e": EventTask,
        "w": WaitTask,
        "h": HoldTask,
        # It was not possible to adjust ifeq to support variables without breaking old
        # macros, so this function is deprecated and if_eq introduced. Kept for backwards
        # compatibility:
        "ifeq": DeprecatedIfEqTask,
    }

    @staticmethod
    def is_this_a_macro(output: Any):
        """Figure out if this is a macro."""
        if not isinstance(output, str):
            return False

        if "+" in output.strip():
            # for example "a + b"
            return True

        return "(" in output and ")" in output and len(output) >= 4

    @staticmethod
    def _extract_args(inner: str):
        """Extract parameters from the inner contents of a call.

        This does not parse them.

        Parameters
        ----------
        inner
            for example '1, r, r(2, k(a))' should result in ['1', 'r', 'r(2, k(a))']
        """
        inner = inner.strip()
        brackets = 0
        params = []
        start = 0
        string = False
        for position, char in enumerate(inner):
            # ignore anything between string quotes
            if char == '"':
                string = not string
            if string:
                continue

            # ignore commas inside child macros
            if char == "(":
                brackets += 1
            if char == ")":
                brackets -= 1
            if char == "," and brackets == 0:
                # , potentially starts another parameter, but only if
                # the current brackets are all closed.
                params.append(inner[start:position].strip())
                # skip the comma
                start = position + 1

        # one last parameter
        params.append(inner[start:].strip())

        return params

    @staticmethod
    def _count_brackets(macro):
        """Find where the first opening bracket closes."""
        openings = macro.count("(")
        closings = macro.count(")")
        if openings != closings:
            raise MacroError(
                macro, f"Found {openings} opening and {closings} closing brackets"
            )

        brackets = 0
        position = 0
        for char in macro:
            position += 1
            if char == "(":
                brackets += 1
                continue

            if char == ")":
                brackets -= 1
                if brackets == 0:
                    # the closing bracket of the call
                    break

        return position

    @staticmethod
    def _split_keyword_arg(param):
        """Split "foo=bar" into "foo" and "bar".

        If not a keyward param, return None and the param.
        """
        if re.match(r"[a-zA-Z_][a-zA-Z_\d]*=.+", param):
            split = param.split("=", 1)
            return split[0], split[1]

        return None, param

    @staticmethod
    def _validate_keyword_argument_names(
        keyword_args: Dict[str, Any],
        task_class: Type[Task],
    ) -> None:
        for keyword_arg in keyword_args:
            for argument in task_class.argument_configs:
                if argument.name == keyword_arg:
                    break
            else:
                raise MacroError(msg=f"Unknown keyword argument {keyword_arg}")

    @staticmethod
    def _parse_recurse(
        code: str,
        context: Optional[Context],
        mapping: Mapping,
        verbose: bool,
        macro_instance: Optional[Macro] = None,
        depth: int = 0,
    ) -> RawValue:
        """Handle a subset of the macro, e.g. one parameter or function call.

        Not using eval for security reasons.

        Parameters
        ----------
        code
            Just like parse. A single parameter or the complete macro as string.
            Comments and redundant whitespace characters are expected to be removed already.
            Example:
            - "parallel(key(a),key(b).key($foo))"
            - "key(a)"
            - "a"
            - "key(b).key($foo)"
            - "b"
            - "key($foo)"
            - "$foo"
        context : Context
        macro_instance
            A macro instance to add tasks to. This is the output of the parser, and is
            organized like a tree.
        depth
            For logging porposes
        """
        assert isinstance(code, str)
        assert isinstance(depth, int)

        def debug(*args, **kwargs):
            if verbose:
                logger.debug(*args, **kwargs)

        space = "  " * depth

        code = code.strip()

        # is it another macro?
        task_call_match = re.match(r"^(\w+)\(", code)
        task_name = task_call_match[1] if task_call_match else None

        if task_name is None:
            # It is probably either a key name like KEY_A or a variable name as in `set(var,1)`,
            # both won't contain special characters that can break macro syntax so they don't
            # have to be wrapped in quotes. The argument configuration of the tasks will
            # detemrine how to parse it.
            debug("%svalue %s", space, code)
            return RawValue(value=code)

        if macro_instance is None:
            # start a new chain
            macro_instance = Macro(code, context, mapping)
        else:
            # chain this call to the existing instance
            assert isinstance(macro_instance, Macro)

        task_class = Parser.TASK_CLASSES.get(task_name)
        if task_class is None:
            raise MacroError(code, f"Unknown function {task_name}")

        # get all the stuff inbetween
        closing_bracket_position = Parser._count_brackets(code) - 1
        inner = code[code.index("(") + 1 : closing_bracket_position]
        debug("%scalls %s with %s", space, task_name, inner)

        # split "3, foo=a(2, k(a).w(10))" into arguments
        raw_string_args = Parser._extract_args(inner)

        # parse and sort the params
        positional_args: List[RawValue] = []
        keyword_args: Dict[str, RawValue] = {}
        for param in raw_string_args:
            key, value = Parser._split_keyword_arg(param)
            parsed = Parser._parse_recurse(
                value.strip(),
                context,
                mapping,
                verbose,
                None,
                depth + 1,
            )
            if key is None:
                if len(keyword_args) > 0:
                    msg = f'Positional argument "{key}" follows keyword argument'
                    raise MacroError(code, msg)
                positional_args.append(parsed)
            else:
                if key in keyword_args:
                    raise MacroError(code, f'The "{key}" argument was specified twice')
                keyword_args[key] = parsed

        debug(
            "%sadd call to %s with %s, %s",
            space,
            task_name,
            positional_args,
            keyword_args,
        )

        Parser._validate_keyword_argument_names(
            keyword_args,
            task_class,
        )
        Parser._validate_num_args(
            code,
            task_name,
            task_class,
            raw_string_args,
        )

        try:
            task = task_class(
                positional_args,
                keyword_args,
                context,
                mapping,
            )
            macro_instance.add_task(task)
        except TypeError as exception:
            raise MacroError(msg=str(exception)) from exception

        # is after this another call? Chain it to the macro_instance
        more_code_exists = len(code) > closing_bracket_position + 1
        if more_code_exists:
            next_char = code[closing_bracket_position + 1]
            statement_closed = next_char == "."

            if statement_closed:
                # skip over the ")."
                chain = code[closing_bracket_position + 2 :]
                debug("%sfollowed by %s", space, chain)
                Parser._parse_recurse(
                    chain,
                    context,
                    mapping,
                    verbose,
                    macro_instance,
                    depth,
                )
            elif re.match(r"[a-zA-Z_]", next_char):
                # something like foo()bar
                raise MacroError(
                    code,
                    f'Expected a "." to follow after '
                    f"{code[:closing_bracket_position + 1]}",
                )

        return RawValue(value=macro_instance)

    @staticmethod
    def _validate_num_args(
        code: str,
        task_name: str,
        task_class: Type[Task],
        raw_string_args: List[str],
    ) -> None:
        min_args, max_args = task_class.get_num_parameters()
        num_provided_args = len(raw_string_args)
        if num_provided_args < min_args or num_provided_args > max_args:
            if min_args != max_args:
                msg = (
                    f"{task_name} takes between {min_args} and {max_args}, "
                    f"not {num_provided_args} parameters"
                )
            else:
                msg = (
                    f"{task_name} takes {min_args}, not {num_provided_args} parameters"
                )

            raise MacroError(code, msg)

    @staticmethod
    def handle_plus_syntax(macro):
        """Transform a + b + c to hold_keys(a,b,c)."""
        if "+" not in macro:
            return macro

        if "(" in macro or ")" in macro:
            raise MacroError(macro, f'Mixing "+" and macros is unsupported: "{ macro}"')

        chunks = [chunk.strip() for chunk in macro.split("+")]

        if "" in chunks:
            raise MacroError(f'Invalid syntax for "{macro}"')

        output = f"hold_keys({','.join(chunks)})"

        logger.debug('Transformed "%s" to "%s"', macro, output)
        return output

    @staticmethod
    def remove_whitespaces(macro, delimiter='"'):
        """Remove whitespaces, tabs, newlines and such outside of string quotes."""
        result = ""
        for i, chunk in enumerate(macro.split(delimiter)):
            # every second chunk is inside string quotes
            if i % 2 == 0:
                result += re.sub(r"\s", "", chunk)
            else:
                result += chunk
            result += delimiter

        # one extra delimiter was added
        return result[: -len(delimiter)]

    @staticmethod
    def remove_comments(macro):
        """Remove comments from the macro and return the resulting code."""
        # keep hashtags inside quotes intact
        result = ""

        for i, line in enumerate(macro.split("\n")):
            for j, chunk in enumerate(line.split('"')):
                if j > 0:
                    # add back the string quote
                    chunk = f'"{chunk}'

                # every second chunk is inside string quotes
                if j % 2 == 0 and "#" in chunk:
                    # everything from now on is a comment and can be ignored
                    result += chunk.split("#")[0]
                    break
                else:
                    result += chunk

            if i < macro.count("\n"):
                result += "\n"

        return result

    @staticmethod
    def clean(code):
        """Remove everything irrelevant for the macro."""
        return Parser.remove_whitespaces(
            Parser.remove_comments(code),
            '"',
        )

    @staticmethod
    def parse(macro: str, context=None, mapping=None, verbose: bool = True) -> Macro:
        """Parse and generate a Macro that can be run as often as you want.

        Parameters
        ----------
        macro
            "repeat(3, key(a).wait(10))"
            "repeat(2, key(a).key(KEY_A)).key(b)"
            "wait(1000).modify(Shift_L, repeat(2, k(a))).wait(10, 20).key(b)"
        context : Context, or None for use in Frontend
        mapping
            the mapping for the macro, or None for use in Frontend
        verbose
            log the parsing True by default
        """
        # TODO pass mapping in frontend and do the target check for keys?
        logger.debug("parsing macro %s", macro.replace("\n", ""))
        macro = Parser.clean(macro)
        macro = Parser.handle_plus_syntax(macro)

        macro_obj = Parser._parse_recurse(
            macro,
            context,
            mapping,
            verbose,
        ).value
        if not isinstance(macro_obj, Macro):
            raise MacroError(macro, "The provided code was not a macro")

        return macro_obj



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


"""See TestEventPipeline for more tests."""

import asyncio
import unittest
from unittest.mock import MagicMock

import evdev
from evdev.ecodes import (
    EV_KEY,
    EV_ABS,
    EV_REL,
    ABS_X,
    REL_X,
    BTN_LEFT,
    BTN_RIGHT,
    KEY_A,
    REL_Y,
    REL_WHEEL,
)

from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.configs.mapping import Mapping, DEFAULT_REL_RATE, KnownUinput
from inputremapper.injection.global_uinputs import GlobalUInputs, UInput
from inputremapper.injection.mapping_handlers.abs_to_abs_handler import AbsToAbsHandler
from inputremapper.injection.mapping_handlers.abs_to_btn_handler import AbsToBtnHandler
from inputremapper.injection.mapping_handlers.abs_to_rel_handler import AbsToRelHandler
from inputremapper.injection.mapping_handlers.axis_switch_handler import (
    AxisSwitchHandler,
)
from inputremapper.injection.mapping_handlers.combination_handler import (
    CombinationHandler,
)
from inputremapper.injection.mapping_handlers.hierarchy_handler import HierarchyHandler
from inputremapper.injection.mapping_handlers.key_handler import KeyHandler
from inputremapper.injection.mapping_handlers.macro_handler import MacroHandler
from inputremapper.injection.mapping_handlers.mapping_handler import (
    MappingHandler,
    InputEventHandler,
)
from inputremapper.injection.mapping_handlers.rel_to_abs_handler import RelToAbsHandler
from inputremapper.injection.mapping_handlers.rel_to_btn_handler import RelToBtnHandler
from inputremapper.injection.mapping_handlers.rel_to_rel_handler import RelToRelHandler
from inputremapper.input_event import InputEvent, EventActions
from tests.lib.cleanup import cleanup
from tests.lib.constants import MAX_ABS
from tests.lib.fixtures import fixtures
from tests.lib.patches import InputDevice
from tests.lib.test_setup import test_setup


class BaseTests:
    """implements test that should pass on most mapping handlers
    in special cases override specific tests.
    """

    handler: MappingHandler

    def setUp(self):
        raise NotImplementedError

    def tearDown(self) -> None:
        cleanup()

    def test_reset(self):
        mock = MagicMock()
        self.handler.set_sub_handler(mock)
        self.handler.reset()
        mock.reset.assert_called()


@test_setup
class TestAxisSwitchHandler(BaseTests, unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        input_combination = InputCombination(
            (
                InputConfig(type=2, code=5),
                InputConfig(type=1, code=3),
            )
        )
        self.global_uinputs = GlobalUInputs(UInput)
        self.global_uinputs.prepare_all()
        self.handler = AxisSwitchHandler(
            input_combination,
            Mapping(
                input_combination=input_combination.to_config(),
                target_uinput="mouse",
                output_type=2,
                output_code=1,
            ),
            MagicMock(),
            self.global_uinputs,
        )


@test_setup
class TestAbsToBtnHandler(BaseTests, unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        input_combination = InputCombination(
            [InputConfig(type=3, code=5, analog_threshold=10)]
        )
        self.global_uinputs = GlobalUInputs(UInput)
        self.global_uinputs.prepare_all()
        self.handler = AbsToBtnHandler(
            input_combination,
            Mapping(
                input_combination=input_combination.to_config(),
                target_uinput="mouse",
                output_symbol="BTN_LEFT",
            ),
            global_uinputs=self.global_uinputs,
        )


@test_setup
class TestAbsToAbsHandler(BaseTests, unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        input_combination = InputCombination([InputConfig(type=EV_ABS, code=ABS_X)])
        self.global_uinputs = GlobalUInputs(UInput)
        self.global_uinputs.prepare_all()
        self.handler = AbsToAbsHandler(
            input_combination,
            Mapping(
                input_combination=input_combination.to_config(),
                target_uinput="gamepad",
                output_type=EV_ABS,
                output_code=ABS_X,
            ),
            global_uinputs=self.global_uinputs,
        )

    async def test_reset(self):
        self.handler.notify(
            InputEvent(0, 0, EV_ABS, ABS_X, MAX_ABS),
            source=InputDevice("/dev/input/event15"),
        )
        self.handler.reset()
        history = self.global_uinputs.get_uinput("gamepad").write_history
        self.assertEqual(
            history,
            [InputEvent.from_tuple((3, 0, MAX_ABS)), InputEvent.from_tuple((3, 0, 0))],
        )


@test_setup
class TestRelToAbsHandler(BaseTests, unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        input_combination = InputCombination([InputConfig(type=EV_REL, code=REL_X)])
        self.global_uinputs = GlobalUInputs(UInput)
        self.global_uinputs.prepare_all()
        self.handler = RelToAbsHandler(
            input_combination,
            Mapping(
                input_combination=input_combination.to_config(),
                target_uinput="gamepad",
                output_type=EV_ABS,
                output_code=ABS_X,
            ),
            self.global_uinputs,
        )

    async def test_reset(self):
        self.handler.notify(
            InputEvent(0, 0, EV_REL, REL_X, 123),
            source=InputDevice("/dev/input/event15"),
        )
        self.handler.reset()
        history = self.global_uinputs.get_uinput("gamepad").write_history
        self.assertEqual(len(history), 2)

        # something large, doesn't matter
        self.assertGreater(history[0].value, MAX_ABS / 10)

        # 0, because of the reset
        self.assertEqual(history[1].value, 0)

    async def test_rate_changes(self):
        expected_rate = 100

        # delta in usec
        delta = 1000000 / expected_rate

        self.handler.notify(
            InputEvent(0, delta, EV_REL, REL_X, 100),
            source=InputDevice("/dev/input/event15"),
        )

        self.handler.notify(
            InputEvent(0, delta * 2, EV_REL, REL_X, 100),
            source=InputDevice("/dev/input/event15"),
        )

        self.assertEqual(self.handler._observed_rate, expected_rate)

    async def test_rate_stays(self):
        # if two timestamps are equal, the rate stays at its previous value,
        # in this case the default

        self.handler.notify(
            InputEvent(0, 50, EV_REL, REL_X, 100),
            source=InputDevice("/dev/input/event15"),
        )

        self.handler.notify(
            InputEvent(0, 50, EV_REL, REL_X, 100),
            source=InputDevice("/dev/input/event15"),
        )

        self.assertEqual(self.handler._observed_rate, DEFAULT_REL_RATE)


@test_setup
class TestAbsToRelHandler(BaseTests, unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        input_combination = InputCombination([InputConfig(type=EV_ABS, code=ABS_X)])
        self.global_uinputs = GlobalUInputs(UInput)
        self.global_uinputs.prepare_all()
        self.handler = AbsToRelHandler(
            input_combination,
            Mapping(
                input_combination=input_combination.to_config(),
                target_uinput="mouse",
                output_type=EV_REL,
                output_code=REL_X,
            ),
            self.global_uinputs,
        )

    async def test_reset(self):
        self.handler.notify(
            InputEvent(0, 0, EV_ABS, ABS_X, MAX_ABS),
            source=InputDevice("/dev/input/event15"),
        )
        await asyncio.sleep(0.2)
        self.handler.reset()
        await asyncio.sleep(0.05)

        count = self.global_uinputs.get_uinput("mouse").write_count
        self.assertGreater(count, 6)  # count should be 60*0.2 = 12
        await asyncio.sleep(0.2)
        self.assertEqual(count, self.global_uinputs.get_uinput("mouse").write_count)


@test_setup
class TestCombinationHandler(BaseTests, unittest.IsolatedAsyncioTestCase):
    handler: CombinationHandler

    def setUp(self):
        mouse = fixtures.foo_device_2_mouse
        self.mouse_hash = mouse.get_device_hash()

        keyboard = fixtures.foo_device_2_keyboard
        self.keyboard_hash = keyboard.get_device_hash()

        gamepad = fixtures.gamepad
        self.gamepad_hash = gamepad.get_device_hash()

        input_combination = InputCombination(
            (
                InputConfig(
                    type=EV_REL,
                    code=5,
                    analog_threshold=10,
                    origin_hash=self.mouse_hash,
                ),
                InputConfig(
                    type=EV_KEY,
                    code=3,
                    origin_hash=self.keyboard_hash,
                ),
                InputConfig(
                    type=EV_KEY,
                    code=4,
                    origin_hash=self.gamepad_hash,
                ),
            )
        )

        self.input_combination = input_combination

        self.context_mock = MagicMock()

        self.global_uinputs = GlobalUInputs(UInput)
        self.global_uinputs.prepare_all()
        self.handler = CombinationHandler(
            input_combination,
            Mapping(
                input_combination=input_combination.to_config(),
                target_uinput="mouse",
                output_symbol="BTN_LEFT",
                release_combination_keys=True,
            ),
            self.context_mock,
            global_uinputs=self.global_uinputs,
        )

        sub_handler_mock = MagicMock(InputEventHandler)
        self.handler.set_sub_handler(sub_handler_mock)

        # insert our own test-uinput to see what is being written to it
        self.uinputs = {
            self.mouse_hash: evdev.UInput(),
            self.keyboard_hash: evdev.UInput(),
            self.gamepad_hash: evdev.UInput(),
        }
        self.context_mock.get_forward_uinput = lambda origin_hash: self.uinputs[
            origin_hash
        ]

    def test_forward_correctly(self):
        # In the past, if a mapping has inputs from two different sub devices, it
        # always failed to send the release events to the correct one.
        # Nowadays, self._context.get_forward_uinput(origin_hash) is used to
        # release them correctly.
        # 1. trigger the combination
        self.handler.notify(
            InputEvent.rel(
                code=self.input_combination[0].code,
                value=1,
                origin_hash=self.input_combination[0].origin_hash,
            ),
            source=fixtures.foo_device_2_mouse,
        )
        self.handler.notify(
            InputEvent.key(
                code=self.input_combination[1].code,
                value=1,
                origin_hash=self.input_combination[1].origin_hash,
            ),
            source=fixtures.foo_device_2_keyboard,
        )
        self.handler.notify(
            InputEvent.key(
                code=self.input_combination[2].code,
                value=1,
                origin_hash=self.input_combination[2].origin_hash,
            ),
            source=fixtures.gamepad,
        )

        # 2. expect release events to be written to the correct devices, as indicated
        # by the origin_hash of the InputConfigs
        self.assertListEqual(
            self.uinputs[self.mouse_hash].write_history,
            [InputEvent.rel(self.input_combination[0].code, 0)],
        )
        self.assertListEqual(
            self.uinputs[self.keyboard_hash].write_history,
            [InputEvent.key(self.input_combination[1].code, 0)],
        )

        # We do not expect a release event for this, because there was no key-down
        # event when the combination triggered.
        # self.assertListEqual(
        #     self.uinputs[self.gamepad_hash].write_history,
        #     [InputEvent.key(self.input_combination[2].code, 0)],
        # )

    def test_no_forwards(self):
        # if a combination is not triggered, nothing is released
        # 1. inject any two events
        self.handler.notify(
            InputEvent.rel(
                code=self.input_combination[0].code,
                value=1,
                origin_hash=self.input_combination[0].origin_hash,
            ),
            source=fixtures.foo_device_2_mouse,
        )
        self.handler.notify(
            InputEvent.key(
                code=self.input_combination[1].code,
                value=1,
                origin_hash=self.input_combination[1].origin_hash,
            ),
            source=fixtures.foo_device_2_keyboard,
        )

        # 2. expect no release events to be written
        self.assertListEqual(self.uinputs[self.mouse_hash].write_history, [])
        self.assertListEqual(self.uinputs[self.keyboard_hash].write_history, [])


@test_setup
class TestHierarchyHandler(BaseTests, unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.mock1 = MagicMock()
        self.mock2 = MagicMock()
        self.mock3 = MagicMock()
        self.global_uinputs = GlobalUInputs(UInput)
        self.global_uinputs.prepare_all()
        self.handler = HierarchyHandler(
            [self.mock1, self.mock2, self.mock3],
            InputConfig(type=EV_KEY, code=KEY_A),
            self.global_uinputs,
        )

    def test_reset(self):
        self.handler.reset()
        self.mock1.reset.assert_called()
        self.mock2.reset.assert_called()
        self.mock3.reset.assert_called()


@test_setup
class TestKeyHandler(BaseTests, unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        input_combination = InputCombination(
            (
                InputConfig(type=2, code=0, analog_threshold=10),
                InputConfig(type=1, code=3),
            )
        )
        self.global_uinputs = GlobalUInputs(UInput)
        self.global_uinputs.prepare_all()
        self.handler = KeyHandler(
            input_combination,
            Mapping(
                input_combination=input_combination.to_config(),
                target_uinput="mouse",
                output_symbol="BTN_LEFT",
            ),
            self.global_uinputs,
        )

    def test_reset(self):
        self.handler.notify(
            InputEvent(0, 0, EV_REL, REL_X, 1, actions=(EventActions.as_key,)),
            source=InputDevice("/dev/input/event11"),
        )
        history = self.global_uinputs.get_uinput("mouse").write_history
        self.assertEqual(history[0], InputEvent.key(BTN_LEFT, 1))
        self.assertEqual(len(history), 1)

        self.handler.reset()
        history = self.global_uinputs.get_uinput("mouse").write_history
        self.assertEqual(history[1], InputEvent.key(BTN_LEFT, 0))
        self.assertEqual(len(history), 2)


@test_setup
class TestMacroHandler(BaseTests, unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.input_combination = InputCombination(
            (
                InputConfig(type=2, code=0, analog_threshold=10),
                InputConfig(type=1, code=3),
            )
        )
        self.context_mock = MagicMock()
        self.global_uinputs = GlobalUInputs(UInput)
        self.global_uinputs.prepare_all()
        self.set_handler(KnownUinput.KEYBOARD, "key(a)")

    def set_handler(self, target_uinput: KnownUinput, macro: str):
        self.handler = MacroHandler(
            self.input_combination,
            Mapping(
                input_combination=self.input_combination.to_config(),
                target_uinput=target_uinput,
                output_symbol=macro,
            ),
            context=self.context_mock,
            global_uinputs=self.global_uinputs,
        )

    async def test_reset(self):
        self.set_handler(KnownUinput.MOUSE, "hold_keys(BTN_LEFT, BTN_RIGHT)")

        self.handler.notify(
            InputEvent(0, 0, EV_REL, REL_X, 1, actions=(EventActions.as_key,)),
            source=InputDevice("/dev/input/event11"),
        )

        await asyncio.sleep(0.1)
        history = self.global_uinputs.get_uinput(KnownUinput.MOUSE).write_history
        self.assertIn(InputEvent.key(BTN_LEFT, 1), history)
        self.assertIn(InputEvent.key(BTN_RIGHT, 1), history)
        self.assertEqual(len(history), 2)

        self.handler.reset()
        await asyncio.sleep(0.1)
        history = self.global_uinputs.get_uinput(KnownUinput.MOUSE).write_history
        self.assertIn(InputEvent.key(BTN_LEFT, 0), history[-2:])
        self.assertIn(InputEvent.key(BTN_RIGHT, 0), history[-2:])
        self.assertEqual(len(history), 4)

    async def test_reset_output(self):
        self.set_handler(KnownUinput.KEYBOARD, "key_down(a)")

        history = self.global_uinputs.get_uinput(KnownUinput.KEYBOARD).write_history

        self.handler.reset()
        await asyncio.sleep(0.1)
        self.assertEqual(len(history), 0)

        self.handler.notify(
            InputEvent(0, 0, EV_REL, REL_X, 1, actions=(EventActions.as_key,)),
            source=InputDevice("/dev/input/event11"),
        )
        await asyncio.sleep(0.1)
        self.assertEqual(len(history), 1)
        self.assertIn(InputEvent.key(KEY_A, 1), history)

        self.handler.reset()
        await asyncio.sleep(0.1)
        self.assertEqual(len(history), 2)
        self.assertIn(InputEvent.key(KEY_A, 0), history)


@test_setup
class TestRelToBtnHandler(BaseTests, unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        input_combination = InputCombination(
            [InputConfig(type=2, code=0, analog_threshold=10)]
        )
        self.global_uinputs = GlobalUInputs(UInput)
        self.global_uinputs.prepare_all()
        self.handler = RelToBtnHandler(
            input_combination,
            Mapping(
                input_combination=input_combination.to_config(),
                target_uinput="mouse",
                output_symbol="BTN_LEFT",
            ),
            self.global_uinputs,
        )


@test_setup
class TestRelToRelHanlder(BaseTests, unittest.IsolatedAsyncioTestCase):
    handler: RelToRelHandler

    def setUp(self):
        input_combination = InputCombination([InputConfig(type=EV_REL, code=REL_X)])
        self.global_uinputs = GlobalUInputs(UInput)
        self.global_uinputs.prepare_all()
        self.handler = RelToRelHandler(
            input_combination,
            Mapping(
                input_combination=input_combination.to_config(),
                output_type=EV_REL,
                output_code=REL_Y,
                output_value=20,
                target_uinput="mouse",
            ),
            self.global_uinputs,
        )

    def test_should_map(self):
        self.assertTrue(
            self.handler._should_map(
                InputEvent(
                    0,
                    0,
                    EV_REL,
                    REL_X,
                    0,
                )
            )
        )
        self.assertFalse(
            self.handler._should_map(
                InputEvent(
                    0,
                    0,
                    EV_REL,
                    REL_WHEEL,
                    1,
                )
            )
        )

    def test_reset(self):
        # nothing special has to happen here
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


import unittest

from evdev.ecodes import (
    EV_REL,
    EV_KEY,
    REL_X,
)

from inputremapper.configs.keyboard_layout import keyboard_layout
from inputremapper.injection.macros.parse import Parser
from tests.lib.test_setup import test_setup
from tests.unit.test_macros.macro_test_base import MacroTestBase, DummyMapping


@test_setup
class TestEvent(MacroTestBase):
    async def test_event_1(self):
        macro = Parser.parse("e(EV_KEY, KEY_A, 1)", self.context, DummyMapping)
        a_code = keyboard_layout.get("a")

        await macro.run(self.handler)
        self.assertListEqual(self.result, [(EV_KEY, a_code, 1)])
        self.assertEqual(self.count_child_macros(macro), 0)

    async def test_event_2(self):
        macro = Parser.parse(
            "repeat(1, event(type=5421, code=324, value=154))",
            self.context,
            DummyMapping,
        )
        code = 324

        await macro.run(self.handler)
        self.assertListEqual(self.result, [(5421, code, 154)])
        self.assertEqual(self.count_child_macros(macro), 1)

    async def test_event_mouse(self):
        macro = Parser.parse("e(EV_REL, REL_X, 10)", self.context, DummyMapping)

        await macro.run(self.handler)
        self.assertListEqual(self.result, [(EV_REL, REL_X, 10)])
        self.assertEqual(self.count_child_macros(macro), 0)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


"""Utility functions."""

import sys
from hashlib import md5
from typing import Optional, NewType

import evdev

DeviceHash = NewType("DeviceHash", str)


def is_service() -> bool:
    return sys.argv[0].endswith("input-remapper-service")


def get_device_hash(device: evdev.InputDevice) -> DeviceHash:
    """get a unique hash for the given device."""
    # The builtin hash() function can not be used because it is randomly
    # seeded at python startup.
    # A non-cryptographic hash would be faster but there is none in the standard lib
    # This hash needs to stay the same across reboots, and even stay the same when
    # moving the config to a new computer.
    s = str(device.capabilities(absinfo=False)) + device.name
    return DeviceHash(md5(s.encode()).hexdigest().lower())


def get_evdev_constant_name(type_: Optional[int], code: Optional[int], *_) -> str:
    """Handy function to get the evdev constant name for display purposes.

    Returns "unknown" for unknown events.
    """
    # using this function is more readable than
    #   type_, code = event.type_and_code
    #   name = evdev.ecodes.bytype[type_][code]
    name = evdev.ecodes.bytype.get(type_, {}).get(code)

    if type(name) in [list, tuple]:
        # python-evdev >= 1.8.0 uses tuples
        name = name[0]

    if name is None:
        return "unknown"

    return name
